# -*- coding: utf-8 -*-

import unittest
from quartzbio.utils import tabulate as t
from quartzbio.utils import printing as p


class TestTabulate(unittest.TestCase):
    def test_classify(self):
        self.assertEqual(t._isnumber("123.45"), True)
        self.assertEqual(t._isnumber("123"), True)
        self.assertEqual(t._isnumber("spam"), False)
        self.assertEqual(t._isint("123"), True)
        self.assertEqual(t._isint("123.45"), False)

        self.assertEqual(t._type(None), t._none_type)
        self.assertEqual(t._type("foo"), t._text_type)
        self.assertEqual(t._type("1"), t._int_type)
        self.assertEqual(t._type(1), t._int_type)
        self.assertEqual(t._type("\x1b[31m42\x1b[0m"), t._int_type)

    def test_align(self):
        self.assertEqual(t._afterpoint("123.45"), 2)
        self.assertEqual(t._afterpoint("1001"), -1)

        self.assertEqual(t._afterpoint("eggs"), -1)
        self.assertEqual(t._afterpoint("123e45"), 2)

        self.assertEqual(t._padleft(6, "abcd"), "  abcd")
        self.assertEqual(t._padright(6, "abcd"), "abcd  ")

        self.assertEqual(t._padboth(6, "abcd"), " abcd ")

        self.assertEqual(t._padboth(7, "abcd"), " abcd  ")

        self.assertEqual(t._padright(2, "abc"), "abc")
        self.assertEqual(t._padleft(2, "abc"), "abc")
        self.assertEqual(t._padboth(2, "abc"), "abc")

        self.assertEqual(
            t._align_column(
                ["12.345", "-1234.5", "1.23", "1234.5", "1e+234", "1.0e234"], "decimal"
            ),
            [
                "   12.345  ",
                "-1234.5    ",
                "    1.23   ",
                " 1234.5    ",
                "    1e+234 ",
                "    1.0e234",
            ],
        )

    def test_column_type(self):
        self.assertEqual(t._column_type(["1", "2"]), t._int_type)
        self.assertEqual(t._column_type(["1", "2.3"]), t._float_type)
        self.assertEqual(t._column_type(["1", "2.3", "four"]), t._text_type)
        self.assertEqual(
            t._column_type(["four", "\u043f\u044f\u0442\u044c"]), t._text_type
        )

        self.assertEqual(t._column_type([None, "brux"]), t._text_type)
        self.assertEqual(t._column_type([1, 2, None]), t._int_type)

    def test_tabulate(self):
        p.TTY_COLS = 80
        tsv = t.simple_separated_format("\t")
        expected = """
foo 	 1
spam\t23
"""
        # [-1:1] below to remove leading and trailing "\n"s above
        got = t.tabulate([["foo", 1], ["spam", 23]], [], tsv)
        self.assertEqual(got, expected[1:-1], "simple separated format table")
        ####################################################################

        expected = """
| abcd   |   12345 |
|--------+---------|
| XY     |       2 |
| lmno   |       4 |
"""
        hrow = ["abcd", "12345"]
        tbl = [["XY", 2], ["lmno", 4]]

        # [-1:1] below to remove leading and trailing "\n"s above
        self.assertEqual(
            t.tabulate(tbl, hrow), expected[1:-1], "org mode with header and unicode"
        )

        ###################################################################

        expected = """
|                Fields | Data        |
|-----------------------+-------------|
|     alternate_alleles | ['T']       |
|       clinical_origin | ['somatic'] |
| clinical_significance | other       |
|          gene_symbols | ['CPB1']    |
"""
        data = [
            ("gene_symbols", ["CPB1"]),
            ("clinical_significance", "other"),
            ("clinical_origin", ["somatic"]),
            ("alternate_alleles", ["T"]),
        ]
        got = t.tabulate(
            data, headers=("Fields", "Data"), aligns=("right", "left"), sort=True
        )

        # [-1:1] below to remove leading and trailing "\n"s above
        self.assertEqual(
            got,
            expected[1:-1],
            "mixed data with arrays; close to actual " + "query output",
        )

        expected = """
|                Fields | Data        |
|-----------------------+-------------|
|          gene_symbols | ['CPB1']    |
| clinical_significance | other       |
|       clinical_origin | ['somatic'] |
|     alternate_alleles | ['T']       |
"""
        got = t.tabulate(
            data, headers=("Fields", "Data"), aligns=("right", "left"), sort=False
        )
        self.assertEqual(got, expected[1:-1], "mixed data with arrays; unsorted")


if __name__ == "__main__":
    unittest.main()



import os
from typing import Any
from typing import Dict
from urllib.parse import urlencode
from urllib.parse import urljoin

import quartzbio

from httpx_oauth.oauth2 import BaseOAuth2


class QuartzBioOAuth2(BaseOAuth2[Dict[str, Any]]):
    """Class implementing OAuth2 for QuartzBio API"""

    # QuartzBio API OAuth2 endpoints
    QUARTZBIO_URL = os.environ.get("QUARTZBIO_URL")
    OAUTH2_TOKEN_URL = "/v1/oauth2/token"
    OAUTH2_REVOKE_TOKEN_URL = "/v1/oauth2/revoke_token"

    def __init__(self, client_id, client_secret, name="quartzbio"):
        if not self.QUARTZBIO_URL:
            raise ValueError("QUARTZBIO_URL environment variable not set")

        super().__init__(
            client_id,
            client_secret,
            self.QUARTZBIO_URL,
            urljoin(quartzbio.get_api_host(), self.OAUTH2_TOKEN_URL),
            revoke_token_endpoint=urljoin(
                quartzbio.get_api_host(), self.OAUTH2_REVOKE_TOKEN_URL
            ),
            name=name,
        )

    def get_authorization_url(self, redirect_uri):
        """Creates authorization url for OAuth2"""

        params = {
            "response_type": "code",
            "client_id": self.client_id,
            "redirect_uri": redirect_uri,
        }

        return "{}/authorize?{}".format(self.authorize_endpoint, urlencode(params))




import uuid

from .helper import QuartzBioTestCase


class APIResourceTests(QuartzBioTestCase):
    def test_apiresource_iteration(self):
        public_vault = self.client.Vault.get_by_full_path("quartzbio:public")
        n_folders = len(list(public_vault.folders(depth=0)))

        folder_iter = public_vault.folders(depth=0)
        for i, j in enumerate(folder_iter):
            pass
        self.assertTrue(i == n_folders - 1)

        # Iterating again should be the same
        for i, j in enumerate(folder_iter):
            pass
        self.assertTrue(i == n_folders - 1)

    def test_apiresource_serialize_metadata(self):
        folder_no_metadata = self.client.Object.get_or_create_by_full_path(
            "~/{}".format(uuid.uuid4()), object_type="folder"
        )

        metadata = folder_no_metadata.metadata
        foo_tuple = ("foo", "bar")

        # foo key-value pair is not in 'metadata' object
        self.assertFalse(foo_tuple in metadata.items())

        metadata["foo"] = "bar"
        params = folder_no_metadata.serialize(folder_no_metadata)

        # Test that foo key-value pair is set as value to 'metadata'
        # key in 'params' dict that will be serialized
        self.assertTrue(foo_tuple in params["metadata"].items())

        folder_with_metadata = self.client.Object.get_or_create_by_full_path(
            "~/{}".format(uuid.uuid4()),
            object_type="folder",
            metadata=dict([foo_tuple]),
        )

        metadata = folder_with_metadata.metadata
        foo_1_tuple = ("foo1", "bar1")

        # foo_1 key-value pair is not in 'metadata' object
        self.assertFalse(foo_1_tuple in metadata.items())

        metadata["foo1"] = "bar1"

        params = folder_with_metadata.serialize(folder_with_metadata)
        params_items = params["metadata"].items()

        # Test that both foo and foo_1 key-value pairs are set as value to 'metadata'
        # key in 'params' dict that will be serialized
        self.assertTrue(foo_tuple in params_items)
        self.assertTrue(foo_1_tuple in params_items)



import time


TIMEOUT = 20  # Seconds


def clean_history(driver, domains):
    temp = driver.get_location()
    for domain in domains:
        driver.open(domain)
        driver.delete_all_visible_cookies()
    driver.open(temp)


def invincible(func):
    def wrap():
        try:
            return func()
        except:
            pass

    return wrap


def switch_windows(driver):
    new_window_handle = None
    while not new_window_handle:
        for handle in driver.window_handles:
            if handle != driver.current_window_handle:
                new_window_handle = handle
                break
    driver.switch_to.window(new_window_handle)
    return new_window_handle


class WaitForTimeout(Exception):
    """This should only be raised inside the `wait_for` function."""

    pass


def wait_for(condition_function, get_message=lambda: "", *args, **kwargs):
    """
    Waits for condition_function to return True or raises WaitForTimeout.
    :param (function) condition_function: Should return True on success.
    :param args: Optional args to pass to condition_function.
    :param kwargs: Optional kwargs to pass to condition_function.
        if `timeout` is in kwargs, it will be used to override TIMEOUT
    :raises: WaitForTimeout If condition_function doesn't return True in time.
    Usage:
        def get_element(selector):
            # some code to get some element or return a `False`-y value.
        selector = '.js-plotly-plot'
        try:
            wait_for(get_element, selector)
        except WaitForTimeout:
            self.fail('element never appeared...')
        plot = get_element(selector)  # we know it exists.
    """

    def wrapped_condition_function():
        """We wrap this to alter the call base on the closure."""
        if args and kwargs:
            return condition_function(*args, **kwargs)
        if args:
            return condition_function(*args)
        if kwargs:
            return condition_function(**kwargs)
        return condition_function()

    if "timeout" in kwargs:
        timeout = kwargs["timeout"]
        del kwargs["timeout"]
    else:
        timeout = TIMEOUT

    start_time = time.time()
    while time.time() < start_time + timeout:
        if wrapped_condition_function():
            return True
        time.sleep(0.5)

    raise WaitForTimeout(get_message())



# -*- coding: utf-8 -*-

from .helper import QuartzBioTestCase


class TestClient(QuartzBioTestCase):
    def test_client_resources(self):
        resources = [
            "Annotator",
            "BatchQuery",
            "Beacon",
            "BeaconSet",
            "Dataset",
            "DatasetCommit",
            "DatasetExport",
            "DatasetField",
            "DatasetImport",
            "DatasetMigration",
            "DatasetTemplate",
            "Expression",
            "Filter",
            "GenomicFilter",
            "Manifest",
            "Object",
            "ObjectCopyTask",
            "Query",
            "Task",
            "User",
            "Vault",
            "VaultSyncTask",
        ]
        for r in resources:
            cls = getattr(self.client, r, None)
            self.assertTrue(cls)
            self.assertEqual(self.client, cls._client)



import mimetypes
import os

COMPRESSIONS = (".gz", ".gzip", ".bz2", ".z", ".zip", ".bgz")


def separate_filename_extension(filename):
    """Separates filename into base name
    and extension while handling compressed
    filename extensions.

    Args:
        filename (str): A filename, can also
            be a full filepath.
    Returns:
        base name (str): Base filename (or path)
            without extension.
        extension (str): File extension beginning
            with leading period (e.g. '.txt').
        compression (str): Compression extension
            with leading period (e.x. '.gz').
    """
    base_filename, file_extension = os.path.splitext(filename)
    if file_extension in COMPRESSIONS and "." in base_filename:
        compression = file_extension
        base_filename, file_extension = os.path.splitext(base_filename)
    else:
        compression = ""
    return base_filename, file_extension, compression


def check_gzip_path(file_path):
    """Check if we have a gzipped file path"""
    _, ftype = mimetypes.guess_type(file_path)
    return ftype == "gzip"


def get_home_dir():
    try:
        # Python 3.5+
        from pathlib import Path

        return str(Path.home())
    except:
        from os.path import expanduser

        return expanduser("~")



# -*- coding: utf-8 -*-

import os
import shutil
import sys
import unittest
import quartzbio
import quartzbio.cli.credentials as creds
import contextlib


@contextlib.contextmanager
def nostdout():
    savestderr = sys.stdout

    class Devnull(object):
        def write(self, _):
            pass

    sys.stdout = Devnull()
    try:
        yield
    finally:
        sys.stdout = savestderr


class TestCredentials(unittest.TestCase):
    def setUp(self):
        self.quartzbiodir = os.path.join(
            os.path.dirname(__file__), "data", ".quartzbio"
        )
        self.api_host = quartzbio.get_api_host()
        quartzbio.client._host = "https://api.quartzbio.com"

    def tearDown(self):
        quartzbio.client._host = self.api_host
        if os.path.isdir(self.quartzbiodir):
            shutil.rmtree(self.quartzbiodir)

    def test_credentials(self):
        datadir = os.path.join(os.path.dirname(__file__), "data")
        os.environ["HOME"] = datadir

        # Make sure we don't have have the test quartzbio directory
        if os.path.isdir(self.quartzbiodir):
            shutil.rmtree(self.quartzbiodir)

        cred_file = creds.netrc.path()
        self.assertTrue(
            os.path.exists(cred_file), "cred file created when it doesn't exist first"
        )

        self.assertEqual(creds.get_credentials(), None, "Should not find credentials")

        test_credentials_file = os.path.join(datadir, "test_creds")
        shutil.copy(test_credentials_file, cred_file)

        auths = creds.get_credentials()
        self.assertTrue(auths is not None, "Should find credentials")

        quartzbio.client._host = "https://example.com"

        auths = creds.get_credentials()
        self.assertEqual(
            auths,
            None,
            f"Should not find credentials for host {quartzbio.get_api_host()}",
        )

        quartzbio.client._host = "https://api.quartzbio.com"
        creds.delete_credentials()
        auths = creds.get_credentials()
        self.assertEqual(
            auths,
            None,
            f"Should not find removed credentials for host {quartzbio.get_api_host()}",
        )

        pair = (
            "testagain@quartzbio.com",
            "b00b00",
        )
        creds.save_credentials(*pair)
        auths = creds.get_credentials()
        self.assertTrue(
            auths is not None,
            "Should get newly set credentials for " "host {0}".format(
                quartzbio.get_api_host()
            ),
        )

        expected = (quartzbio.get_api_host(), pair[0], "Token", pair[1])
        self.assertEqual(auths, expected, "Should get back creds we saved")



# -*- coding: utf-8 -*-

from .client import client
from .resource import Object
from .resource import Vault
from .query import QueryBase
from .query import Query
from .query import Filter

import logging

logger = logging.getLogger("quartzbio")


class GlobalSearch(Query):
    """
    GlobalSearch acts as a request wrapper that generates a request from Filter objects,
    and can iterate through streaming result sets.
    """

    def __init__(
        self,
        query=None,
        filters=None,
        entities=None,
        entities_match="any",
        vault_scope="all",
        ordering=None,
        limit=float("inf"),
        page_size=QueryBase.DEFAULT_PAGE_SIZE,
        result_class=dict,
        debug=False,
        raw_results=False,
        **kwargs,
    ):
        """
        Creates a new Query object.

        :Parameters:
          - `query` (optional): An optional query string (advanced search).
          - `filters` (optional): Filter or List of filter objects.
          - `entities` (optional): List of entity tuples to filter on (entity type, entity).
          - `entities_match` (optional): Can be 'all' or 'any' (match any provided entity).
          - `vault_scope` (optional): Can be 'all' or 'access'.
          - `ordering` (optional): List of fields to order the results by.
          - `limit` (optional): Maximum number of query results to return.
          - `page_size` (optional): Number of results to fetch per query page.
          - `result_class` (optional): Class of object returned by query.
          - `debug` (optional): Sends debug information to the API.
          - `raw_results` (optional): Whether to use raw API response or to cast logical
             objects to Vault and Object instances.
        """
        super(GlobalSearch, self).__init__(None)
        self._data_url = "/v2/search"
        self._query = query
        self._entities = entities
        self._entities_match = entities_match
        self._vault_scope = vault_scope
        self._ordering = ordering
        self._result_class = result_class
        self._debug = debug
        self._raw_results = raw_results
        self._error = None

        if filters:
            if isinstance(filters, Filter):
                filters = [filters]
        else:
            filters = []
        self._filters = filters

        # init response and cursor
        self._response = None
        # Limit defines the total number of results that will be returned
        # from a query involving 1 or more pagination requests.
        self._limit = limit
        # Page size/offset are the low level API limit and offset params.
        self._page_size = int(page_size)
        # Page offset can only be set by execute(). It is always set to the
        # current absolute offset contained in the buffer.
        self._page_offset = None
        # slice is set when the Query is being sliced.
        # In this case, __iter__() and next() will not
        # reset the page_offset to 0 before iterating.
        self._slice = None

        # parameter error checking
        if self._limit < 0:
            raise Exception("'limit' parameter must be >= 0")

        if not 0 < self._page_size <= self.MAX_PAGE_SIZE:
            raise Exception(
                "'page_size' parameter must be in " "range [1, {}]".format(
                    self.MAX_PAGE_SIZE
                )
            )

        # Set up the QuartzBioClient
        # (kwargs overrides pre-set, which overrides global)
        self._client = kwargs.get("client") or self._client or client

    def _clone(self, filters=None, entities=None, limit=None):
        new = self.__class__(
            query=self._query,
            limit=self._limit,
            entities=self._entities,
            ordering=self._ordering,
            page_size=self._page_size,
            result_class=self._result_class,
            vault_scope=self._vault_scope,
            entities_match=self._entities_match,
            debug=self._debug,
            client=self._client,
        )
        new._filters += self._filters

        if entities:
            new._entities = entities

        if filters:
            new._filters += filters

        if limit is not None:
            new._limit = limit

        return new

    def __len__(self):
        """
        Returns the total number of results returned in a query. It is the
        number of items you can iterate over.

        In contrast to count(), the result does take into account any limit
        given. In SQL it is like:

              SELECT COUNT(*) FROM (
                 SELECT * FROM <table> [WHERE condition] [LIMIT number]
              )
        """
        return super(GlobalSearch, self).__len__()

    def _build_query(self, **kwargs):
        q = {}

        if self._query:
            q["query"] = self._query

        if self._filters:
            filters = self._process_filters(self._filters)
            if len(filters) > 1:
                q["filters"] = [{"and": filters}]
            else:
                q["filters"] = filters

        if self._entities is not None:
            q["entities"] = self._entities

        if self._ordering is not None:
            q["ordering"] = self._ordering

        if self._vault_scope is not None:
            q["vault_scope"] = self._vault_scope

        if self._entities_match is not None:
            q["entities_match"] = self._entities_match

        if self._debug:
            q["debug"] = "True"

        # Add or modify query parameters
        # (used by BatchQuery and facets)
        q.update(**kwargs)

        return q

    def execute(self, offset=0, **query):
        def _process_result(result):
            # Internally the client uses object_type, not type
            result["object_type"] = result["type"]
            if result["object_type"] == "vault":
                return Vault.construct_from(result)
            else:
                return Object.construct_from(result)

        # Call superclass method execute
        super(GlobalSearch, self).execute(offset, **query)

        # Cast logical objects from response to Object/Vault instances
        if not self._raw_results:
            self._response["results"] = [
                _process_result(i) for i in self._response["results"]
            ]

    def entity(self, **kwargs):
        """
        Returns GlobalSearch instance with the query args combined with
        existing set with AND.

        kwargs can contain only one entity, entity_type as parameter name and entity as its value.
        If entity is already set for the GlobalSearch, it will be overridden.
        """

        if not kwargs:
            raise AttributeError("Faceting requires at least one field")

        return self._clone(entities=list(kwargs.items()))

    def subjects(self):
        """Returns the list of subjects"""

        # Executes a query to get a full API response which contains subjects list
        gs = self.limit(0)
        gs.execute(include_subjects=True, include_all_subjects=True)

        return gs._response.get("subjects")

    def subjects_count(self):
        """Returns the number of subjects"""

        # Executes a query to get a full API response which contains subjects list
        gs = self.limit(0)
        gs.execute()

        return gs._response.get("subjects_count")

    def vaults(self):
        """Returns the list of vaults"""

        # Executes a query to get a full API response which contains vaults list
        gs = self.limit(0)
        gs.execute(include_vaults=True)

        return gs._response.get("vaults")




from .helper import QuartzBioTestCase


class BeaconTests(QuartzBioTestCase):
    # TEST_DATASET_FULL_PATH = 'quartzbio:public:/ClinVar/3.7.4-2017-01-30/Variants-GRCh37'  # noqa
    TEST_DATASET_FULL_PATH = "quartzbio:Public:/ClinVar/5.2.0-20210110/Variants-GRCH38"

    def test_beacon_request(self):
        """
        Check that current Clinvar/Variants returns correct
        fields for beacon
        """
        dataset = self.client.Object.get_by_full_path(self.TEST_DATASET_FULL_PATH)
        beacon = dataset.beacon(
            chromosome="6",
            coordinate=51612854,
            allele="G",  # staging
        )

        check_fields = ["query", "exist", "total"]

        for f in check_fields:
            self.assertTrue(f in beacon)

        # Check that Clinvar/Variants version 3.7.0-2015-12-06
        # returns true for specific case

        dataset = self.client.Object.get_by_full_path(self.TEST_DATASET_FULL_PATH)
        beacontwo = dataset.beacon(chromosome="13", coordinate=113803460, allele="T")

        self.assertTrue(beacontwo["exist"])
        self.assertEqual(beacontwo["total"], 1)



import mock

from quartzbio.test.client_mocks import fake_export_create

from .helper import QuartzBioTestCase


class TestDatasetExports(QuartzBioTestCase):
    def _validate_export(self, export, dataset, **kwargs):
        self.assertEqual(export.dataset_id, dataset.id)

    @mock.patch("quartzbio.resource.DatasetExport.create")
    def test_export_from_query(self, Create):
        Create.side_effect = fake_export_create

        # Test with params
        params = {
            "fields": ["my_field"],
            "limit": 100,
        }
        target_fields = [dict(name="test")]

        dataset = self.client.Dataset(1)
        export = dataset.export(
            params=params,
            dataset=dataset,
            target_fields=target_fields,
            target_full_path="~/hello",
            format="tsv.gz",
            follow=False,
        )
        self.assertEqual(export.dataset_id, dataset.id)
        self.assertEqual(export.target_fields, target_fields)
        self.assertEqual(export.target_full_path, "~/hello")
        self.assertEqual(export.format, "tsv.gz")
        for k in ["fields", "limit"]:
            self.assertEqual(export.params[k], params[k])




import unittest

from .helper import QuartzBioTestCase


class VaultTests(QuartzBioTestCase):
    def test_vaults(self):
        vaults = self.client.Vault.all()
        vault = vaults.data[0]
        self.assertTrue("id" in vault, "Should be able to get id in vault")

        vault2 = self.client.Vault.retrieve(vault.id)
        self.assertEqual(
            vault, vault2, "Retrieving vault id {0} found by all()".format(vault.id)
        )

        check_fields = [
            "account_id",
            "created_at",
            "description",
            "has_children",
            "has_folder_children",
            "id",
            "is_deleted",
            "is_public",
            "last_synced",
            "name",
            "permissions",
            "provider",
            "require_unique_paths",
            "updated_at",
            "url",
            "user_id",
            "vault_properties",
            "vault_type",
        ]

        for f in check_fields:
            self.assertTrue(f in vault, "{0} field is present".format(f))

    def test_vault_paths(self):
        user = self.client.User.retrieve()
        domain = user.account.domain
        user_vault = "user-{0}".format(user.id)

        vaults = self.client.Vault.all()
        for vault in vaults:
            v, v_paths = self.client.Vault.validate_full_path(vault.full_path)
            self.assertEqual(v, vault.full_path)

        test_cases = [
            ["myVault/", "{0}:myVault".format(domain)],
            ["myVault", "{0}:myVault".format(domain)],
            ["{0}:myVault".format(domain), "{0}:myVault".format(domain)],
            ["acme:myVault", "acme:myVault"],
            ["myVault/folder1/folder2: xyz", "{0}:myVault".format(domain)],
            ["acme:myVault/folder1/folder2: xyz", "acme:myVault"],
            ["acme:myVault:/folder1/folder2: xyz", "acme:myVault"],
            # The following are the "new" vault/path formats:
            ["~/", "{0}:{1}".format(domain, user_vault)],
            ["acme:myVault/uploads_folder", "acme:myVault"],
            ["myVault/uploads_folder", "{0}:myVault".format(domain)],
        ]
        for case, expected in test_cases:
            v, v_paths = self.client.Vault.validate_full_path(case)
            self.assertEqual(v, expected)

        error_test_cases = [
            "",
            "/",
            ":",
            ":/",
            "::/",
            "x:",
            # Underscore in domain
            "my_Domain:myVault:/the/heack",
            # Space in domain
            "my Domain:my:Vault:/the/heack",
            # Too many colons
            "myDomain:my:Vault:/the/heack",
            "oops:myDomain:myVault",
        ]
        for case in error_test_cases:
            with self.assertRaises(Exception):
                v, v_paths = self.client.Vault.validate_full_path(case)

    @unittest.skip("Skip because API Host on GH pipelines doesn't support versioning.")
    def test_vault_versioning(self):
        vault = self.client.Vault.get_personal_vault()
        initial_versioning_status = vault["versioning"]
        assert initial_versioning_status in ["enabled", "disabled", "suspended"]

        self.addCleanup(
            VaultTests.clean_up_after_vault_versioning, vault, initial_versioning_status
        )

        vault.enable_versioning()
        assert "enabled" == vault["versioning"]
        vault.disable_versioning()
        assert "disabled" == vault["versioning"]
        vault.suspend_versioning()
        assert "suspended" == vault["versioning"]

    @staticmethod
    def clean_up_after_vault_versioning(vault, versioning_status):
        if versioning_status == "enabled":
            vault.enable_versioning()
        elif versioning_status == "disabled":
            vault.disable_versioning()
        elif versioning_status == "suspended":
            vault.suspend_versioning()
        assert versioning_status == vault["versioning"]




from .apiresource import ListObject
from .user import User
from .dataset import Dataset
from .datasetfield import DatasetField
from .datasetimport import DatasetImport
from .datasetexport import DatasetExport
from .datasetcommit import DatasetCommit
from .datasetmigration import DatasetMigration
from .datasettemplate import DatasetTemplate
from .vault_sync_task import VaultSyncTask
from .object_copy_task import ObjectCopyTask
from .dataset_restore_task import DatasetRestoreTask
from .dataset_snapshot_task import DatasetSnapshotTask
from .manifest import Manifest
from .object import Object
from .vault import Vault
from .task import Task
from .beacon import Beacon
from .beaconset import BeaconSet
from .application import Application
from .group import Group
from .savedquery import SavedQuery


types = {
    "Application": Application,
    "Beacon": Beacon,
    "BeaconSet": BeaconSet,
    "Dataset": Dataset,
    "DatasetImport": DatasetImport,
    "DatasetExport": DatasetExport,
    "DatasetCommit": DatasetCommit,
    "DatasetMigration": DatasetMigration,
    "DatasetTemplate": DatasetTemplate,
    "DatasetField": DatasetField,
    "DatasetRestoreTask": DatasetRestoreTask,
    "DatasetSnapshotTask": DatasetSnapshotTask,
    "Group": Group,
    "Manifest": Manifest,
    "Object": Object,
    "ObjectCopyTask": ObjectCopyTask,
    "ECSTask": Task,
    "VaultSyncTask": VaultSyncTask,
    "User": User,
    "Vault": Vault,
    "list": ListObject,
    "SavedQuery": SavedQuery,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
import copy
import argparse

import quartzbio
from quartzbio import QuartzBioError

from . import auth
from . import data
from .tutorial import print_tutorial
from .ipython import launch_ipython_shell
from ..auth import validate_api_host_url
from ..utils.files import get_home_dir


class TildeFixStoreAction(argparse._StoreAction):
    """A special "store" action for argparse that replaces
    any detected home directory with a tilde.
    (reverses bash's built-in ~ expansion).
    """

    def __call__(self, parser, namespace, values, option_string=None):
        home = get_home_dir()
        if values and values.startswith(home):
            values = values.replace(home, "~", 1)
        setattr(namespace, self.dest, values)


# KEY=VALUE argparser
# https://stackoverflow.com/a/56521375/305633
class KeyValueDictAppendAction(argparse.Action):
    """
    argparse action to split an argument into KEY=VALUE form
    on the first = and append to a dictionary.
    """

    def __call__(self, parser, args, values, option_string=None):
        assert len(values) == 1
        try:
            (k, v) = values[0].split("=", 2)
        except ValueError:
            raise argparse.ArgumentError(
                self, "could not parse argument '{}' as k=v format".format(values[0])
            )
        d = getattr(args, self.dest) or {}
        d[k] = v
        setattr(args, self.dest, d)


class QuartzBioArgumentParser(argparse.ArgumentParser):
    """
    Main parser for the QuartzBio command line client.
    """

    subcommands = {
        "login": {
            "func": auth.login_and_save_credentials,
            "help": "Login and save credentials",
            "arguments": [
                {
                    "flags": "--debug",
                    "action": "store_true",
                    "default": False,
                    "help": "Shows the source of the user credentials",
                },
            ]
        },
        "logout": {"func": auth.logout, "help": "Logout and delete saved credentials"},
        "whoami": {"func": auth.whoami, "help": "Show your QuartzBio email address"},
        "tutorial": {
            "func": print_tutorial,
            "help": "Show the QuartzBio Python Tutorial",
        },
        "shell": {
            "func": launch_ipython_shell,
            "help": "Open the QuartzBio Python shell",
        },
        "import": {
            "func": data.import_file,
            "help": "Import a local file into a QuartzBio dataset",
            "arguments": [
                {
                    "flags": "--create-vault",
                    "action": "store_true",
                    "help": "Create the vault if it doesn't exist",
                },
                {
                    "flags": "--create-dataset",
                    "action": "store_true",
                    "help": "Create the dataset if it doesn't exist",
                },
                {
                    "flags": "--capacity",
                    "default": "small",
                    "help": "Specifies the capacity of the created dataset: "
                    "small (default, <100M records), "
                    "medium (<500M), large (>=500M)",
                },
                {
                    "name": "--tag",
                    "help": "A tag to be added. "
                    "Tags are case insensitive strings. Example tags: "
                    '--tag GRCh38 --tag Tissue --tag "Foundation Medicine"',
                    "action": "append",
                },
                {
                    "name": "--metadata",
                    "help": "Dataset metadata in the format KEY=VALUE ",
                    "nargs": 1,
                    "metavar": "KEY=VALUE",
                    "action": KeyValueDictAppendAction,
                },
                {
                    "name": "--metadata-json-file",
                    "help": "Metadata key value pairs in JSON format",
                },
                {
                    "flags": "--template-id",
                    "help": "The template ID used when "
                    "creating a new dataset (via --create-dataset)",
                },
                {
                    "flags": "--template-file",
                    "help": "A local template file to be used when "
                    "creating a new dataset (via --create-dataset)",
                },
                {
                    "flags": "--follow",
                    "action": "store_true",
                    "default": False,
                    "help": "Follow the import's progress until it completes",
                },
                {
                    "flags": "--commit-mode",
                    "default": "append",
                    "help": "Commit mode to use when importing data. "
                    'Options are "append" (default), "overwrite",'
                    '"upsert", or "delete"',
                },
                {
                    "flags": "--remote-source",
                    "action": "store_true",
                    "default": False,
                    "help": "File paths are remote globs or full paths on "
                    "the QuartzBio file system.",
                },
                {
                    "flags": "--dry-run",
                    "help": "Dry run mode will not create any datasets or "
                    "import any files.",
                    "action": "store_true",
                },
                {
                    "name": "full_path",
                    "help": "The full path to the dataset in the format: "
                    '"domain:vault:/path/dataset". ',
                    "action": TildeFixStoreAction,
                },
                {
                    "name": "file",
                    "help": "One or more files to import. Can be local files, "
                    "folders, globs or remote URLs. Pass --remote-source in "
                    "order to list remote full_paths or path globs on the "
                    "QuartzBio file system.",
                    "nargs": "+",
                },
            ],
        },
        "create-dataset": {
            "func": data.create_dataset,
            "help": "Create a QuartzBio dataset",
            "arguments": [
                {
                    "flags": "--create-vault",
                    "action": "store_true",
                    "help": "Create the vault if it doesn't exist",
                },
                {
                    "flags": "--template-id",
                    "help": "The template ID used when "
                    "creating a new dataset (via --create-dataset)",
                },
                {
                    "flags": "--template-file",
                    "help": "A local template file to be used when "
                    "creating a new dataset (via --create-dataset)",
                },
                {
                    "flags": "--capacity",
                    "default": "small",
                    "help": "Specifies the capacity of the dataset: "
                    "small (default, <100M records), "
                    "medium (<500M), large (>=500M)",
                },
                {
                    "name": "--tag",
                    "help": "A tag to be added. "
                    "Tags are case insensitive strings. Example tags: "
                    '--tag GRCh38 --tag Tissue --tag "Foundation Medicine"',
                    "action": "append",
                },
                {
                    "name": "--metadata",
                    "help": "Dataset metadata in the format KEY=VALUE ",
                    "nargs": 1,
                    "metavar": "KEY=VALUE",
                    "action": KeyValueDictAppendAction,
                },
                {
                    "name": "--metadata-json-file",
                    "help": "Metadata key value pairs in JSON format",
                },
                {
                    "flags": "--dry-run",
                    "help": "Dry run mode will not create the dataset",
                    "action": "store_true",
                },
                {
                    "name": "full_path",
                    "help": "The full path to the dataset in the format: "
                    '"domain:vault:/path/dataset". '
                    "Defaults to your personal vault if no vault is provided. "
                    "Defaults to the vault root if no path is provided.",
                    "action": TildeFixStoreAction,
                },
            ],
        },
        "upload": {
            "func": data.upload,
            "help": "Upload a file or directory to a QuartzBio Vault",
            "arguments": [
                {
                    "flags": "--full-path",
                    "help": "The full path where the files and folders should "
                    "be created, defaults to the root of your personal vault",
                    "action": TildeFixStoreAction,
                    "default": "~/",
                },
                {
                    "flags": "--num-processes",
                    "help": "Number of uploads to process in parallel. Defaults "
                    "to 1, but can be set much higher than CPU count since the "
                    "upload process is IO bound, not CPU bound.",
                    "default": 1,
                    "type": int,
                },
                {
                    "flags": "--create-full-path",
                    "help": "Creates --full-path location if it does "
                    "not exist. NOTE: This will not create new vaults.",
                    "action": "store_true",
                },
                {
                    "flags": "--exclude",
                    "help": "Paths to files or folder to be excluded from "
                    "upload. Unix shell-style wildcards are supported.",
                    "action": "append",
                },
                {
                    "flags": "--dry-run",
                    "help": "Dry run mode will not upload any files or "
                    "create any folders.",
                    "action": "store_true",
                },
                {
                    "flags": "--archive-folder",
                    "help": "Path to archive files that already exist. "
                    "If a folder is supplied, instead of overwriting "
                    "or creating an incremented filename, the original "
                    "remote file will be moved to this archive folder "
                    "with a timestamp.",
                    "action": TildeFixStoreAction,
                },
                {
                    "flags": "--follow-shortcuts",
                    "help": "Resolves shortcuts when Uploading.",
                    "action": "store_true",
                },
                {
                    "flags": "--max-retries",
                    "help": "Maximum number of retries per upload part for multipart uploads. "
                    "Defaults to 3.",
                    "default": 3,
                    "type": int,
                },
                {
                    "name": "local_path",
                    "help": "The path to the local file or directory " "to upload",
                    "nargs": "+",
                },
            ],
        },
        "ls": {
            "func": data.ls,
            "help": "List files and directories in a QuartzBio Vault",
            "arguments": [
                {
                    "flags": "--recursive",
                    "help": "Recursively list the contents of subdirectories.",
                    "action": "store_true",
                },
                {
                    "flags": "--follow-shortcuts",
                    "help": "Resolves shortcuts when listing.",
                    "action": "store_true",
                },
                {
                    "name": "full_path",
                    "help": "The full path where the files and folders should "
                    "be listed from, defaults to the root of your personal vault",
                    "action": TildeFixStoreAction,
                    "default": "~/",
                },
            ],
        },
        "download": {
            "func": data.download,
            "help": "Download one or more files from a QuartzBio Vault.",
            "arguments": [
                {
                    "flags": "--dry-run",
                    "help": "Dry run mode will not download any files or "
                    "create any folders. Use this mode before using the "
                    "--delete flag.",
                    "action": "store_true",
                },
                {
                    "flags": "full_path",
                    "help": "The full path to the files on QuartzBio. Supports "
                    "Unix style globs in order to download multiple files. "
                    "Note: Downloads are not recursive unless --recursive "
                    "flag is used.",
                    "action": TildeFixStoreAction,
                },
                {
                    "name": "local_path",
                    "help": "The path to the local directory where "
                    "to download files.",
                },
                {
                    "flags": "--recursive",
                    "help": "Downloads files recursively. Note that empty "
                    "folders will be ignored.",
                    "action": "store_true",
                },
                {
                    "flags": "--exclude",
                    "help": "Pattern to match against local full paths "
                    "of files to be excluded from downloading. "
                    "This pattern is only used when --recursive is used. "
                    "Unix shell-style wildcards are supported. "
                    "Exclude patterns will always be superseded by include "
                    "patterns.",
                    "action": "append",
                },
                {
                    "flags": "--include",
                    "help": "Pattern to match against full paths "
                    "of files to be included for downloading. "
                    "This pattern is only used when --recursive is used. "
                    "Unix shell-style wildcards are supported. "
                    "Include patterns will always supersede exclude "
                    "patterns.",
                    "action": "append",
                },
                {
                    "flags": "--delete",
                    "help": "Deletes local files not found in remote full path. "
                    "Warning, this is dangerous and will delete any files found in "
                    'local path. Do not use a top-level local path such as "/" and '
                    "always use the --dry-run mode to evaluate any changes. "
                    "Empty folders will be deleted.",
                    "action": "store_true",
                },
                {
                    "flags": "--follow-shortcuts",
                    "help": "Resolves shortcuts when downloading. "
                    "If a shortcut to a file is found the target file will be downloaded "
                    "under the shortcut name.",
                    "action": "store_true",
                },
                {
                    "flags": "--num-processes",
                    "help": "Number of downloads to process in parallel. "
                    "If not specified downloads won't be executed in parallel."
                    "If a number less than 1 is set defaults to the number of system CPUs.",
                    "default": None,
                    "type": int,
                },
            ],
        },
        "tag": {
            "func": data.tag,
            "help": "Apply tags or remove tags on objects",
            "arguments": [
                {
                    "flags": "full_path",
                    "help": "The full path of the files, "
                    "folders or datasets to apply the tag updates. "
                    "Unix shell-style wildcards are supported. ",
                    "nargs": "+",
                },
                {
                    "name": "--tag",
                    "help": "A tag to be added/removed. "
                    "Files, folders and datasets can be tagged. "
                    "Tags are case insensitive strings. Example tags: "
                    '--tag GRCh38 --tag Tissue --tag "Foundation Medicine"',
                    "action": "append",
                    "required": True,
                },
                {
                    "flags": "--remove",
                    "help": "Will remove tags instead of adding them.",
                    "action": "store_true",
                },
                {
                    "flags": "--exclude",
                    "help": "Paths to files or folder to be excluded from "
                    "tagging. Unix shell-style wildcards are supported.",
                    "action": "append",
                },
                {
                    "flags": "--tag-folders-only",
                    "help": "Will only apply tags to folders (tags "
                    "all objects by default). ",
                    "action": "store_true",
                },
                {
                    "flags": "--tag-files-only",
                    "help": "Will only apply tags to files (tags "
                    "all objects by default). ",
                    "action": "store_true",
                },
                {
                    "flags": "--tag-datasets-only",
                    "help": "Will only apply tags to datasets (tags "
                    "all objects by default). ",
                    "action": "store_true",
                },
                {
                    "flags": "--dry-run",
                    "help": "Dry run mode will not save tags.",
                    "action": "store_true",
                },
                {
                    "flags": "--no-input",
                    "help": "Automatically accept changes (overrides " "user prompt)",
                    "action": "store_true",
                },
            ],
        },
        "queue": {
            "func": data.show_queue,
            "help": "Shows the current job queue, grouped by User",
        },
    }

    def __init__(self, *args, **kwargs):
        super(QuartzBioArgumentParser, self).__init__(*args, **kwargs)
        self._optionals.title = "QuartzBio Options"
        self.add_argument(
            "--version", action="version", version=quartzbio.version.VERSION
        )
        self.add_argument(
            "--api-host",
            help="Override the default QuartzBio API host",
            type=self.api_host_url,
        )
        self.add_argument("--api-key", help="Manually provide a QuartzBio API key")
        self.add_argument(
            "--access-token", help="Manually provide a QuartzBio OAuth2 access token"
        )

    def _add_subcommands(self):
        """
        The _add_subcommands method must be separate from the __init__
        method, as infinite recursion will occur otherwise, due to the fact
        that the __init__ method itself will be called when instantiating
        a subparser, as we do below
        """
        subcmd_params = {"title": "QuartzBio Commands", "dest": "subcommands"}
        subcmd = self.add_subparsers(**subcmd_params)  # pylint: disable=star-args

        subcommands = copy.deepcopy(self.subcommands)
        for name, params in subcommands.items():
            p = subcmd.add_parser(name, help=params["help"])
            p.set_defaults(func=params["func"])
            for arg in params.get("arguments", []):
                name_or_flags = arg.pop("name", None) or arg.pop("flags", None)
                p.add_argument(name_or_flags, **arg)

    def parse_quartzbio_args(self, args=None, namespace=None):
        """
        Try to parse the args first, and then add the subparsers. We want
        to do this so that we can check to see if there are any unknown
        args. We can assume that if, by this point, there are no unknown
        args, we can append shell to the unknown args as a default.
        However, to do this, we have to suppress stdout/stderr during the
        initial parsing, in case the user calls the help method (in which
        case we want to add the additional arguments and *then* call the
        help method. This is a hack to get around the fact that argparse
        doesn't allow default subcommands.
        """
        try:
            sys.stdout = sys.stderr = open(os.devnull, "w")
            _, unknown_args = self.parse_known_args(args, namespace)
            if not unknown_args:
                args.insert(0, "shell")
        except SystemExit:
            pass
        finally:
            sys.stdout.flush()
            sys.stderr.flush()
            sys.stdout, sys.stderr = sys.__stdout__, sys.__stderr__
        self._add_subcommands()
        return super(QuartzBioArgumentParser, self).parse_args(args, namespace)

    def api_host_url(self, value):
        if not value:
            raise QuartzBioError("No QuartzBio API host is set")

        validate_api_host_url(value)
        return value


def main():

    """Main entry point for QuartzBio CLI"""
    parser = QuartzBioArgumentParser()
    args = parser.parse_quartzbio_args(sys.argv[1:])

    quartzbio.login(
        api_host=args.api_host,
        api_key=args.api_key,
        access_token=args.access_token,
    )

    return args.func(args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from urllib.parse import unquote

import os
import requests
import tempfile
import shutil

try:
    from urllib import quote_plus
except ImportError:
    from urllib.parse import quote_plus

from ..client import client, _handle_api_error, _handle_request_error
from ..utils.tabulate import tabulate
from ..utils.printing import pager

# from quartzbio.errors import NotFoundError
from ..errors import NotFoundError

from .util import class_to_api_name
from .quartzbio_object import QuartzBioObject, convert_to_quartzbio_object


class APIResource(QuartzBioObject):
    """Abstract Class for an API Resource"""

    @classmethod
    def retrieve(cls, id, **params):
        instance = cls(id, **params)
        instance.refresh()
        return instance

    def refresh(self):
        self.refresh_from(self.request("get", self.instance_url()))
        return self

    @classmethod
    def class_name(cls):
        if cls == APIResource:
            raise NotImplementedError(
                "APIResource is an abstract class.  You should perform "
                "actions on its subclasses (e.g. Vault, Object, Dataset)"
            )
        return str(quote_plus(cls.__name__))

    @classmethod
    def class_url(cls):
        """Returns a versioned URI string for this class"""
        base = "v{0}".format(getattr(cls, "RESOURCE_VERSION", "1"))
        return "/{0}/{1}".format(base, class_to_api_name(cls.class_name()))

    def instance_url(self):
        """Get instance URL by ID"""
        id_ = self.get(self.ID_ATTR)
        base = self.class_url()

        if id_:
            return "/".join([base, str(id_)])
        else:
            raise Exception(
                "Could not determine which URL to request: %s instance "
                "has invalid ID: %r" % (type(self).__name__, id_),
                self.ID_ATTR,
            )


class ListObject(QuartzBioObject):
    def all(self, **params):
        """Lists all items in a class that you have access to"""
        return self.request("get", self["url"], params=params)

    def create(self, **params):
        return self.request("post", self["url"], data=params)

    def first_page(self, **params):
        url = getattr(self, "_first_page_url", self["url"])
        return self.request("get", url, params=params)

    def next_page(self, **params):
        if self["links"]["next"]:
            return self.request("get", self["links"]["next"], params=params)
        return None

    def prev_page(self, **params):
        if self["links"]["prev"]:
            return self.request("get", self["links"]["prev"], params=params)
        return None

    def quartzbio_objects(self):
        return convert_to_quartzbio_object(self["data"], client=self._client)

    def set_tabulate(self, fields, **kwargs):
        self._tabulate = lambda data: tabulate(
            [[d[i] for i in fields] for d in data], **kwargs
        )

    def __len__(self):
        return self["total"]

    def __str__(self):
        if getattr(self, "_tabulate", None):
            return "\n" + self._tabulate(self["data"])
        return super(ListObject, self).__str__()

    def __iter__(self):
        self._i = 0

        # Store the URL of the first page if we haven't yet.
        self._first_page_url = getattr(self, "_first_page_url", self["url"])
        if self._first_page_url != self["url"]:
            self.refresh_from(self.first_page())

        return self

    def __next__(self):
        """Python 3"""
        return self.next()

    def next(self):
        if not getattr(self, "_i", None):
            self._i = 0

        if self._i >= len(self["data"]):
            # get the next page of results
            next_page = self.next_page()
            if next_page is None:
                raise StopIteration
            self.refresh_from(next_page)
            self._i = 0

        obj = convert_to_quartzbio_object(self["data"][self._i], client=self._client)
        self._i += 1
        return obj


class SingletonAPIResource(APIResource):
    @classmethod
    def retrieve(cls, **kwargs):
        _client = kwargs.pop("client", None) or cls._client or client
        return super(SingletonAPIResource, cls).retrieve(None, client=_client)

    @classmethod
    def class_url(cls):
        """
        Returns a versioned URI string for this class,
        and don't pluralize the class name.
        """
        base = "v{0}".format(getattr(cls, "RESOURCE_VERSION", "1"))
        return "/{0}/{1}".format(
            base, class_to_api_name(cls.class_name(), pluralize=False)
        )

    def instance_url(self):
        return self.class_url()


class CreateableAPIResource(APIResource):
    @classmethod
    def create(cls, **params):
        _client = params.pop("client", None) or cls._client or client
        url = cls.class_url()
        response = _client.post(url, data=params)
        return convert_to_quartzbio_object(response, client=_client)


class DeletableAPIResource(APIResource):
    def delete(self, **params):
        printable_name = class_to_api_name(self.class_name(), pluralize=False).replace(
            "_", " "
        )
        if not params.pop("force", False):
            res = input(
                "Are you sure you want to delete this %s? " "[y/N] " % printable_name
            )
            if res.strip().lower() != "y":
                print("Not performing deletion.")
                return

        response = self.request("delete", self.instance_url(), params=params)
        return convert_to_quartzbio_object(response, client=self._client)


class DownloadableAPIResource(APIResource):
    def download(self, path=None, **kwargs):
        """
        Download the file to the specified directory or file path.
        Downloads to a temporary directory if no path is specified.

        Returns the absolute path to the file.
        """
        download_url = self.download_url(**kwargs)
        try:
            # For vault objects, use the object's filename
            # as the fallback if none is specified.
            filename = self.filename
        except AttributeError:
            # If the object has no filename attribute,
            # extract one from the download URL.
            filename = download_url.split("%3B%20filename%3D")[1]
            # Remove additional URL params from the name and "unquote" it.
            filename = unquote(filename.split("&")[0])

        if path:
            path = os.path.expanduser(path)
            # If the path is a dir, use the extracted filename
            if os.path.isdir(path):
                path = os.path.join(path, filename)
        else:
            # Create a temporary directory for the file
            path = os.path.join(tempfile.gettempdir(), filename)

        try:
            # use streaming to prevent automatic decompression
            response = requests.request(method='get', url=download_url, stream=True)
        except Exception as e:
            _handle_request_error(e)

        if not (200 <= response.status_code < 400):
            _handle_api_error(response)

        with open(path, "wb") as fileobj:
            if filename.endswith('.gz'):
                # Don't automatically decompress gzipped files
                shutil.copyfileobj(response.raw, fileobj)
            else:
                for chunk in response.iter_content(chunk_size=1024 * 8):
                    if chunk:
                        fileobj.write(chunk)

        return path

    def download_url(self, **kwargs):
        download_url = self.instance_url()
        if "version_id" in kwargs:
            download_url += "/versions/{}".format(kwargs["version_id"])
        else:
            download_url += "/download"
        # Don't redirect, just return the signed S3 URL
        kwargs.update({"redirect": ""})
        response = self.request(
            "get", download_url, params=kwargs, allow_redirects=False
        )

        return response.get("url") or response.get("download_url")


class ListableAPIResource(APIResource):
    """Has one method: *all()* which lists everything in the resource."""

    @classmethod
    def all(cls, **params):
        _client = params.pop("client", None) or cls._client or client
        url = cls.class_url()
        response = _client.get(url, params)
        results = convert_to_quartzbio_object(response, client=_client)

        # If the object has LIST_FIELDS, setup tabulate
        if len(results.data) > 0:
            list_fields = getattr(results.data[0], "LIST_FIELDS", None)
            if list_fields:
                fields, headers = list(zip(*list_fields))
                results.set_tabulate(fields, headers=headers, sort=False)
        return results

    @classmethod
    def _retrieve_helper(cls, model_name, field_name, error_value, **params):
        _client = params.pop("client", None) or cls._client or client
        url = cls.class_url()
        response = _client.get(url, params)
        results = convert_to_quartzbio_object(response, client=_client)
        objects = results.data
        allow_multiple = params.pop("allow_multiple", None)
        if len(objects) > 1:
            if allow_multiple:
                return objects
            else:
                raise Exception(
                    'Multiple {0}s found with {1} "{2}"'.format(
                        model_name, field_name, error_value
                    )
                )
        elif len(objects) == 1:
            return objects[0]
        else:
            raise NotFoundError(
                'No {0} found with {1} "{2}"'.format(
                    model_name, field_name, error_value
                )
            )

    @classmethod
    def pager(cls, **params):
        return pager(cls.all, **params)

    def __repr__(self):
        return tabulate(
            list(self.items()), ["Fields", "Data"], aligns=["right", "left"], sort=True
        )


class SearchableAPIResource(APIResource):
    @classmethod
    def search(cls, query="", **params):
        _client = params.pop("client", None) or cls._client or client
        params.update({"q": query})
        url = cls.class_url()
        response = _client.get(url, params)
        results = convert_to_quartzbio_object(response, client=_client)

        # If the object has LIST_FIELDS, setup tabulate
        if len(results.data) > 0:
            list_fields = getattr(results.data[0], "LIST_FIELDS", None)
            if list_fields:
                fields, headers = list(zip(*list_fields))
                results.set_tabulate(fields, headers=headers)

        return results


class UpdateableAPIResource(APIResource):
    def save(self):
        self.refresh_from(
            self.request("patch", self.instance_url(), data=self.serialize(self))
        )
        return self

    def serialize(self, obj):
        params = {}
        unsaved_values = obj._unsaved_values

        # adding 'metadata' attribute to be serialized
        if getattr(obj, "metadata", None) is not None:
            unsaved_values.add("metadata")

        for k in unsaved_values:
            if k == "id" or k == self.ID_ATTR:
                continue

            attr = getattr(obj, k)

            if isinstance(attr, QuartzBioObject):
                params[k] = dict(attr)
            else:
                params[k] = attr

        return params



import sys
import os


def _print(msg):
    """
    Handle Python 2 interactive shells without requiring
    new print() format.
    """
    sys.stdout.write(msg + "\n")


def launch_ipython_shell(args):  # pylint: disable=unused-argument
    """Open the QuartzBio shell (IPython wrapper)"""
    try:
        import IPython  # noqa
    except ImportError:
        _print(
            "The QuartzBio Python shell requires IPython.\n"
            "To install, type: 'pip install ipython'"
        )
        return False

    if hasattr(IPython, "version_info"):
        if IPython.version_info > (5, 0, 0, ""):
            return launch_ipython_5_shell(args)

    _print(
        "WARNING: Please upgrade IPython (you are running version: {})".format(
            IPython.__version__
        )
    )
    return launch_ipython_legacy_shell(args)


def launch_ipython_5_shell(args):
    """Open the QuartzBio shell (IPython wrapper) with IPython 5+"""
    import IPython  # noqa
    from traitlets.config import Config

    c = Config()
    path = os.path.dirname(os.path.abspath(__file__))

    try:
        # see if we're already inside IPython
        get_ipython  # pylint: disable=undefined-variable
        _print("WARNING: Running IPython within IPython.")
    except NameError:
        c.InteractiveShell.banner1 = "QuartzBio Python shell started.\n"

    c.InteractiveShellApp.exec_files = ["{}/ipython_init.py".format(path)]
    IPython.start_ipython(argv=[], config=c)


def launch_ipython_legacy_shell(args):  # pylint: disable=unused-argument
    """Open the QuartzBio shell (IPython wrapper) for older IPython versions"""
    try:
        from IPython.config.loader import Config
    except ImportError:
        _print(
            "The QuartzBio Python shell requires IPython.\n"
            "To install, type: 'pip install ipython'"
        )
        return False

    try:
        # see if we're already inside IPython
        get_ipython  # pylint: disable=undefined-variable
    except NameError:
        cfg = Config()
        prompt_config = cfg.PromptManager
        prompt_config.in_template = "[QuartzBio] In <\\#>: "
        prompt_config.in2_template = "   .\\D.: "
        prompt_config.out_template = "Out<\\#>: "
        banner1 = "\nQuartzBio Python shell started."

        exit_msg = "Quitting QuartzBio shell."
    else:
        _print("Running nested copies of IPython.")
        cfg = Config()
        banner1 = exit_msg = ""

    # First import the embeddable shell class
    try:
        from IPython.terminal.embed import InteractiveShellEmbed
    except ImportError:
        # pylint: disable=import-error,no-name-in-module
        from IPython.frontend.terminal.embed import InteractiveShellEmbed

    path = os.path.dirname(os.path.abspath(__file__))
    init_file = "{}/ipython_init.py".format(path)
    exec(compile(open(init_file).read(), init_file, "exec"), globals(), locals())

    InteractiveShellEmbed(config=cfg, banner1=banner1, exit_msg=exit_msg)()



import logging

logger = logging.getLogger("quartzbio")


class NotFoundError(Exception):
    pass


class FileUploadError(Exception):
    pass


class QuartzBioError(Exception):
    """Exceptions tailored to the kinds of errors from a QuartzBio API
    request"""

    default_message = (
        "Unexpected error communicating with QuartzBio. "
        "If this problem persists, let us know at "
        "support@quartzbio.com."
    )

    def __init__(self, message=None, response=None):
        self.json_body = {}
        self.status_code = None
        self.message = message or self.default_message
        self.field_errors = []
        self.response = response

        if response is not None:
            self.status_code = response.status_code
            try:
                self.json_body = response.json()
                logger.debug(
                    "API Response (%d): %s" % (self.status_code, self.json_body)
                )
            except:
                logger.debug(
                    "API Response (%d): Response does not contain JSON."
                    % self.status_code
                )

            if self.status_code == 400:
                self.message = "Bad Request ({})".format(response.url)
            elif response.status_code == 401:
                self.message = "401 Unauthorized ({})".format(response.url)
            elif response.status_code == 403:
                self.message = "403 Forbidden ({})".format(response.url)
            elif response.status_code == 404:
                self.message = "404 Not Found ({})".format(response.url)

            # Handle other keys
            for k, v in list(self.json_body.items()):
                if k in ["non_field_errors", "detail"]:
                    self.message += "\nError: "
                else:
                    self.message += "\nError (%s): " % k

                # can be a list, dict, string
                self.message += str(v)

    def __str__(self):
        return self.message



"""QuartzBio ObjectCopyTask API Resource"""

import time
from .apiresource import ListableAPIResource
from .apiresource import CreateableAPIResource
from .apiresource import UpdateableAPIResource
from .task import Task


class ObjectCopyTask(CreateableAPIResource, ListableAPIResource, UpdateableAPIResource):
    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("status", "Status"),
        ("source_vault_id", "Source Vault"),
        ("target_vault_id", "Target Vault"),
        ("source_object_id", "Source"),
        ("target_object_id", "Target"),
        ("created_at", "Created"),
    )

    def follow(self, loop=True, sleep_seconds=Task.SLEEP_WAIT_DEFAULT):
        if self.status == "queued":
            print("Waiting for Object Copy task (id = {0}) to start...".format(self.id))

        _status = self.status
        while self.status in ["queued", "running"]:
            if self.status != _status:
                print(
                    "Object Copy task is now {0} (was {1})".format(self.status, _status)
                )
                _status = self.status

            if self.status == "running":
                print("Object Copy task '{0}' is {1}".format(self.id, self.status))

            if not loop:
                return

            time.sleep(sleep_seconds)
            self.refresh()

        if self.status == "completed":
            print("Object Copy task complete!")



from quartzbio import Filter
from quartzbio.test.helper import QuartzBioTestCase


class BaseFileQueryTest(QuartzBioTestCase):
    """Test File Queries"""

    def setUp(self):
        super(BaseFileQueryTest, self).setUp()
        self.file = self.client.Object.get_by_full_path(self.TEST_FILE_FULL_PATH)

    def test_basic(self):
        results = self.file.query().filter(
            hgnc_id__in=["HGNC:5", "HGNC:37133", "HGNC:24086"]
        )

        self.assertEqual(results.total, 3)
        self.assertEqual(len(results), results.total)

        # Test that iteration returns the correct number of results.
        # Test iterating through result-sets that are smaller than
        # the page size.
        self.assertEqual(len(results), len([r for r in results]))

    def test_basic_with_limit(self):
        limit = 10
        results = self.file.query(limit=limit)
        self.assertEqual(len(results), limit)

        # test that iteration returns the correct number of results.
        self.assertEqual(len([r for r in results]), limit)

    def test_count(self):
        q = self.file.query()
        total = q.count()
        self.assertGreater(total, 0)

        q = self.file.query().filter(hgnc_id="HGNC:5")
        self.assertEqual(q.count(), 1)

        # invalid filter
        q = self.file.query().filter(hgnc_id="foo")
        self.assertEqual(q.count(), 0)

    def test_count_with_limit(self):
        q = self.file.query()
        total = q.count()
        self.assertGreater(total, 0)

        for limit in [1, 10, 1000]:
            q = self.file.query(limit=limit).filter(hgnc_id="HGNC:5")
            self.assertEqual(q.count(), 1)

            # invalid filter
            q = self.file.query(limit=limit).filter(hgnc_id="foo")
            self.assertEqual(q.count(), 0)

    def test_len(self):
        q = self.file.query()
        total = q.count()
        self.assertGreater(total, 0)
        self.assertEqual(len(q), total)

        q = self.file.query().filter(hgnc_id="HGNC:5")
        self.assertEqual(len(q), 1)

        # invalid filter
        q = self.file.query().filter(hgnc_id="foo")
        self.assertEqual(len(q), 0)

    def test_len_with_limit(self):
        q = self.file.query()
        total = q.count()
        self.assertGreater(total, 0)
        self.assertEqual(len(q), total)

        for limit in [1, 10, 1000]:
            q = self.file.query(limit=limit).filter(hgnc_id="HGNC:5")
            self.assertEqual(len(q), 1 if limit > 0 else 0)

            # invalid filter
            q = self.file.query(limit=limit).filter(hgnc_id="foo")
            self.assertEqual(len(q), 0)

    def test_empty(self):
        # invalid filter
        results = self.file.query().filter(hgnc_id="foo")
        self.assertEqual(len(results), 0)
        self.assertEqual(results[:]._buffer, [])
        self.assertRaises(IndexError, lambda: results[0])

    def test_empty_with_limit(self):
        limit = 100

        # invalid filter
        results = self.file.query(limit=limit).filter(hgnc_id="foo")
        self.assertEqual(len(results), 0)
        self.assertEqual(results[:]._buffer, [])
        self.assertRaises(IndexError, lambda: results[0])

    def test_paging(self):
        page_size = 10
        num_pages = 3
        results = self.file.query(page_size=page_size)

        _results = []
        for i, r in enumerate(results):
            # fetch three pages and break
            if i / page_size == num_pages:
                break
            _results.append(r)

        self.assertEqual(len(_results), num_pages * page_size)
        self.assertEqual(len(set(map(str, _results))), num_pages * page_size)

    def test_paging_with_limit(self):
        page_size = 10
        num_pages = 3
        limit = num_pages * page_size - 1
        results = self.file.query(limit=limit, page_size=page_size)

        _results = []
        for i, r in enumerate(results):
            _results.append(r)

        self.assertEqual(len(_results), limit)
        self.assertEqual(len(set(map(str, _results))), limit)

    def test_slice_ranges(self):
        limit = 50

        results = self.file.query(limit=limit)
        self.assertEqual(len(results[0:limit]), limit)

        results = self.file.query(limit=limit)
        self.assertEqual(len(results[:limit]), limit)

        results = self.file.query(limit=limit)
        self.assertEqual(len(results[limit:]), limit)

        r0 = self.file.query(limit=limit)[0:limit][limit - 1]
        r1 = self.file.query(limit=limit)[limit - 1 :][0]
        self.assertEqual(r0["entrez_id"], r1["entrez_id"])

    def test_slice_ranges_with_paging(self):
        limit = 50
        page_size = 10

        results = self.file.query(limit=limit, page_size=page_size)
        self.assertEqual(len(results[0:limit]), limit)

        results = self.file.query(limit=limit, page_size=page_size)
        self.assertEqual(len(results[:limit]), limit)

        results = self.file.query(limit=limit, page_size=page_size)
        self.assertEqual(len(results[limit:]), limit)

        r0 = self.file.query(limit=limit)[0:limit][limit - 1]
        r1 = self.file.query(limit=limit)[limit - 1 :][0]
        self.assertEqual(r0["entrez_id"], r1["entrez_id"])

    def test_slice_offsets(self):
        zero_two = self.file.query()[0:2]
        one_three = self.file.query()[1:3]

        # Ensure that the repr for [0:1] != [1:2]
        self.assertNotEqual(repr(zero_two), repr(one_three))

        # Ensure that the second repr for [0:2] == [1:3]
        self.assertEqual(repr(zero_two[1]), repr(one_three[0]))

    def test_slice_until_object_end(self):
        total = self.file.query(limit=1000).count()

        sliced_ds = self.file.query(limit=10)[total - 1 :]

        left_records = 0
        for _ in sliced_ds:
            left_records += 1

        # Ensure that only one records has left in the file
        self.assertEqual(left_records, 1)

    def test_slice_ranges_with_small_limit(self):
        # Test slices larger than 'limit'

        limit = 1
        filters = Filter(**{"entrez_id__in": [str(i) for i in range(2, 13)]})
        results = self.file.query(limit=limit, filters=filters)[0:4]
        self.assertEqual(len(results), limit)

    def test_paging_and_slice_equivalence(self):
        idx0 = 3
        idx1 = 5
        filters = Filter(**{"entrez_id__in": [str(i) for i in range(2, 1000)]})

        def _query():
            return self.file.query(limit=10, filters=filters)

        results_slice = _query()[idx0:idx1]
        results_paging = []

        for i, r in enumerate(_query()):
            if i == idx1:
                break
            elif i >= idx0:
                results_paging.append(r)

        self.assertEqual(len(results_paging), len(results_slice))

    def test_field_filters(self):
        limit = 1
        results = self.file.query(limit=limit)
        self.assertEqual(len(results[0].keys()), 52)

        results = self.file.query(limit=limit, fields=["entrez_id"])
        self.assertEqual(len(results[0].keys()), 1)

        results = self.file.query(limit=limit, exclude_fields=["entrez_id"])
        self.assertEqual(len(results[0].keys()), 51)
        self.assertTrue("entrez_id" not in results[0].keys())




from .helper import QuartzBioTestCase


class DatasetTests(QuartzBioTestCase):
    """
    Test Dataset, DatasetField, and Facets
    """

    def test_dataset_retrieval(self):
        dataset = self.client.Dataset.get_by_full_path(self.TEST_DATASET_FULL_PATH)
        self.assertTrue("id" in dataset, "Should be able to get id in dataset")

        check_fields = [
            "class_name",
            "created_at",
            "data_url",
            "vault_id",
            "vault_object_id",
            "description",
            "fields_url",
            "id",
            "updated_at",
            "url",
            "documents_count",
        ]

        for f in check_fields:
            self.assertTrue(f in dataset, "{0} field is present".format(f))

    def test_dataset_tree_traversal_shortcuts(self):
        dataset = self.client.Dataset.get_by_full_path(self.TEST_DATASET_FULL_PATH)

        # get vault object
        self.assertEqual(dataset.vault_object.full_path, self.TEST_DATASET_FULL_PATH)

        # get vault object parent
        self.assertEqual(
            dataset.vault_object.parent.full_path,
            "/".join(self.TEST_DATASET_FULL_PATH.split("/")[:-1]),
        )

        # get vault
        self.assertEqual(
            dataset.vault_object.vault.full_path,
            ":".join(self.TEST_DATASET_FULL_PATH.split(":")[:-1]),
        )

    def test_dataset_fields(self):
        dataset = self.client.Object.get_by_full_path(self.TEST_DATASET_FULL_PATH)
        fields = dataset.fields()
        dataset_field = fields.data[0]
        self.assertTrue(
            "id" in dataset_field, "Should be able to get id in list of dataset fields"
        )

        check_fields = set(
            [
                "class_name",
                "created_at",
                "data_type",
                "dataset_id",
                "title",
                "description",
                "facets_url",
                "ordering",
                "is_hidden",
                "is_valid",
                "is_list",
                "entity_type",
                "expression",
                "name",
                "updated_at",
                "is_read_only",
                "depends_on",
                "id",
                "url",
                "vault_id",
                "url_template",
            ]
        )
        self.assertSetEqual(set(dataset_field.keys()), check_fields)

    def test_dataset_facets(self):
        dataset = self.client.Object.get_by_full_path(self.TEST_DATASET_FULL_PATH)
        field = dataset.fields("status")
        facets = field.facets()
        self.assertTrue(len(facets["facets"]) >= 0)



# -*- coding: utf-8 -*-
import re

try:
    import json
except ImportError:
    json = None

# test for compatible json module
if not (json and hasattr(json, "loads")):
    import simplejson as json  # noqa


def camelcase_to_underscore(name):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def class_to_api_name(name, pluralize=True):
    if pluralize:
        if name.endswith("y"):
            name = name[:-1] + "ie"
        name = name + "s"

    return camelcase_to_underscore(name)



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

import sys
import warnings

VERSION = 'undefined'
install_requires = ['pyprind', 'requests>=2.0.0', 'urllib3>=1.26.0']
extra = {}

with open('quartzbio/version.py') as f:
    for row in f.readlines():
        if row.startswith('VERSION'):
            exec(row)

# quartzbio-recipes requires additional packages
recipes_requires = [
    'pyyaml==5.3.1',
    'click==7.1.2',
    'ruamel.yaml==0.16.12'
]
extras_requires = {
    "recipes": recipes_requires
}

with open('README.md') as f:
    long_description = f.read()

setup(
    name='quartzbio',
    version=VERSION,
    description='The QuartzBio Python client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.8',
    author='Precision for Medicine, Inc.',
    author_email='qb-help@precisionformedicine.com',
    url='https://github.com/quartzbio/quartzbio-python',
    packages=find_packages(),
    package_dir={'quartzbio': 'quartzbio', 'recipes': 'recipes'},
    test_suite='quartzbio.test',
    include_package_data=True,
    install_requires=install_requires,
    platforms='any',
    extras_require=extras_requires,
    entry_points={
        'console_scripts': ['quartzbio = quartzbio.cli.main:main',
                            'quartzbio-recipes = recipes.sync_recipes:sync_recipes']
    },
    classifiers=[
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    **extra
)




import unittest

from quartzbio.query import Filter
from quartzbio import QuartzBioError

from .helper import QuartzBioTestCase


class BaseQueryTest(QuartzBioTestCase):
    """Test Paging Queries"""

    def setUp(self):
        super(BaseQueryTest, self).setUp()
        self.dataset = self.client.Object.get_by_full_path(self.TEST_DATASET_FULL_PATH)
        self.dataset2 = self.client.Object.get_by_full_path(
            self.TEST_DATASET_FULL_PATH_2
        )

    def test_basic(self):
        results = self.dataset.query().filter(
            omim_id__in=["OMIM:123631", "OMIM:123670", "OMIM:123690", "OMIM:306250"]
        )
        self.assertEqual(results.total, 4)
        self.assertEqual(len(results), results.total)

        # Test that iteration returns the correct number of results.
        # Test iterating through result-sets that are smaller than
        # the page size.
        self.assertEqual(len(results), len([r for r in results]))

    def test_basic_with_limit(self):
        limit = 10
        results = self.dataset.query(limit=limit)
        self.assertEqual(len(results), limit)
        self.assertRaises(IndexError, lambda: results[results.total + 1])

        # test that iteration returns the correct number of results.
        self.assertEqual(len([r for r in results]), limit)

    def test_count(self):
        q = self.dataset.query()
        total = q.count()
        self.assertGreater(total, 0)

        # with a filter
        q = self.dataset.query().filter(omim_id="OMIM:123631")
        self.assertEqual(q.count(), 1)

        # with a bogus filter
        q = self.dataset.query().filter(omim_id="OMIM:99999")
        self.assertEqual(q.count(), 0)

    def test_count_with_limit(self):
        q = self.dataset.query()
        total = q.count()
        self.assertGreater(total, 0)

        for limit in [0, 10, 1000]:
            # with a filter
            q = self.dataset.query(limit=limit).filter(omim_id="OMIM:123631")
            self.assertEqual(q.count(), 1)

            # with a bogus filter
            q = self.dataset.query(limit=limit).filter(omim_id="OMIM:99999")
            self.assertEqual(q.count(), 0)

    def test_len(self):
        q = self.dataset.query()
        total = q.count()
        self.assertGreater(total, 0)
        self.assertEqual(len(q), total)

        # with a filter
        q = self.dataset.query().filter(omim_id="OMIM:123631")
        self.assertEqual(len(q), 1)

        # with a bogus filter
        q = self.dataset.query().filter(omim_id="OMIM:999999")
        self.assertEqual(len(q), 0)

    def test_len_with_limit(self):
        q = self.dataset.query()
        total = q.count()
        self.assertGreater(total, 0)
        self.assertEqual(len(q), total)

        for limit in [0, 10, 1000]:
            # with a filter
            q = self.dataset.query(limit=limit).filter(omim_id="OMIM:123631")
            self.assertEqual(len(q), 1 if limit > 0 else 0)

            # with a bogus filter
            q = self.dataset.query(limit=limit).filter(omim_id="OMIM:999999")
            self.assertEqual(len(q), 0)

    def test_empty(self):
        """
        test Query when limit is specified and is GREATER THAN total available
        results.
        """
        # bogus filter
        results = self.dataset.query().filter(omim_id="OMIM:99999")
        self.assertEqual(len(results), 0)
        self.assertEqual(results[:]._buffer, [])
        self.assertRaises(IndexError, lambda: results[0])

    def test_empty_with_limit(self):
        """
        test Query when limit is specified and is GREATER THAN total available
        results.
        """
        limit = 100
        # bogus filter
        results = self.dataset.query(limit=limit).filter(omim_id="OMIM:99999")
        self.assertEqual(len(results), 0)
        self.assertEqual(results[:]._buffer, [])
        self.assertRaises(IndexError, lambda: results[0])

    def test_filter(self):
        """
        test Filtered Query in which limit is specified but is GREATER THAN
        the number of total available results
        """
        num_filters = 4
        filters = (
            Filter(omim_id="OMIM:123631")
            | Filter(omim_id="OMIM:123670")
            | Filter(omim_id="OMIM:123690")
            | Filter(omim_id="OMIM:306250")
        )
        results = self.dataset.query(filters=filters)
        self.assertEqual(len(results), num_filters)
        self.assertRaises(IndexError, lambda: results[num_filters])

    def test_filter_with_limit(self):
        """
        test Filtered Query in which limit is specified but is GREATER THAN
        the number of total available results
        """
        limit = 10
        num_filters = 4
        filters = (
            Filter(omim_id="OMIM:123631")
            | Filter(omim_id="OMIM:123670")
            | Filter(omim_id="OMIM:123690")
            | Filter(omim_id="OMIM:306250")
        )
        results = self.dataset.query(limit=limit, filters=filters)
        self.assertEqual(len(results), num_filters)
        self.assertRaises(IndexError, lambda: results[num_filters])

    def test_paging(self):
        page_size = 10
        num_pages = 3
        results = self.dataset.query(page_size=page_size)

        _results = []
        for i, r in enumerate(results):
            # fetch three pages and break
            if i / page_size == num_pages:
                break
            _results.append(r)

        self.assertEqual(len(_results), num_pages * page_size)
        self.assertEqual(len(set(map(str, _results))), num_pages * page_size)

    def test_paging_with_limit(self):
        page_size = 10
        num_pages = 3
        limit = num_pages * page_size - 1
        results = self.dataset.query(limit=limit, page_size=page_size)

        _results = []
        for i, r in enumerate(results):
            _results.append(r)

        self.assertEqual(len(_results), limit)
        self.assertEqual(len(set(map(str, _results))), limit)

    def test_slice_ranges(self):
        limit = 50

        results = self.dataset.query(limit=limit)
        self.assertEqual(len(results[0:limit]), limit)

        results = self.dataset.query(limit=limit)
        self.assertEqual(len(results[:limit]), limit)

        results = self.dataset.query(limit=limit)
        self.assertEqual(len(results[limit:]), limit)

        r0 = self.dataset.query(limit=limit)[0:limit][limit - 1]
        r1 = self.dataset.query(limit=limit)[limit - 1 :][0]
        self.assertEqual(r0["entrez_id"], r1["entrez_id"])

    def test_slice_ranges_with_paging(self):
        limit = 50
        page_size = 10

        results = self.dataset.query(limit=limit, page_size=page_size)
        self.assertEqual(len(results[0:limit]), limit)

        results = self.dataset.query(limit=limit, page_size=page_size)
        self.assertEqual(len(results[:limit]), limit)

        results = self.dataset.query(limit=limit, page_size=page_size)
        self.assertEqual(len(results[limit:]), limit)

        r0 = self.dataset.query(limit=limit)[0:limit][limit - 1]
        r1 = self.dataset.query(limit=limit)[limit - 1 :][0]
        self.assertEqual(r0["entrez_id"], r1["entrez_id"])

    def test_slice_offsets(self):
        zero_two = self.dataset.query()[0:2]
        one_three = self.dataset.query()[1:3]

        # Ensure that the repr for [0:1] != [1:2]
        self.assertNotEqual(repr(zero_two), repr(one_three))

        # Ensure that the second repr for [0:2] == [1:3]
        self.assertEqual(repr(zero_two[1]), repr(one_three[0]))

    def test_slice_until_object_end(self):
        total = self.dataset.query(limit=0).count()

        sliced_ds = self.dataset.query(limit=10)[total - 1 :]

        left_records = 0
        for _ in sliced_ds:
            left_records += 1

        # Ensure that only one records has left in the dataset
        self.assertEqual(left_records, 1)

    def test_slice_ranges_with_small_limit(self):
        # Test slices larger than 'limit'
        # query returns 6
        limit = 1
        results = self.dataset.query(limit=limit).filter(mamit_trnadb__range=(2, 12))[
            0:4
        ]
        self.assertEqual(len(results), limit)

    def test_paging_and_slice_equivalence(self):
        idx0 = 3
        idx1 = 5

        def _query():
            return self.dataset.query(limit=10).filter(mamit_trnadb__range=(2, 12))

        results_slice = _query()[idx0:idx1]
        results_paging = []

        for i, r in enumerate(_query()):
            if i == idx1:
                break
            elif i >= idx0:
                results_paging.append(r)

        self.assertEqual(len(results_paging), len(results_slice))

        for i in range(0, len(results_slice)):
            id_a = results_paging[i]["entrez_id"]
            id_b = results_slice[i]["entrez_id"]
            self.assertEqual(id_a, id_b)

    def test_caching(self):
        idx0 = 60
        idx1 = 81

        q = self.dataset.query(limit=100)
        results_slice = q[idx0:idx1]
        results_cached = q[idx0:idx1]
        self.assertEqual(len(results_slice), len(results_cached))
        for i in range(0, len(results_slice)):
            id_a = results_slice[i]["status"]
            id_b = results_cached[i]["status"]
            self.assertEqual(id_a, id_b)

    def test_get_by_index(self):
        limit = 100
        page_size = 10
        idxs = [0, 1, 10, 20, 50, 99]
        q = self.dataset.query(limit=limit, page_size=page_size)
        cached = []
        for idx in idxs:
            cached.append(q[idx])

        # forwards
        for i, idx in enumerate(idxs):
            self.assertEqual(cached[i], q[idx])

        # backwards
        for i, idx in reversed(list(enumerate(idxs))):
            self.assertEqual(cached[i], q[idx])

    def test_field_filters(self):
        limit = 1
        results = self.dataset.query(limit=limit)
        self.assertEqual(len(results[0].keys()), 55)

        results = self.dataset.query(limit=limit, fields=["entrez_id"])
        self.assertEqual(len(results[0].keys()), 1)

        results = self.dataset.query(limit=limit, exclude_fields=["entrez_id"])
        self.assertEqual(len(results[0].keys()), 54)
        self.assertTrue("entrez_id" not in results[0].keys())

    def test_entity_filters(self):
        entities = [("gene", "BRCA2")]
        query = self.dataset.query(entities=entities)
        self.assertEqual(query.count(), 1)

    def test_target_fields(self):
        # Invalid fields
        try:
            query = self.dataset.query(target_fields=[1])
        except QuartzBioError as e:
            assert "{u'target_fields': [u'Invalid data']}" in e
            pass

        entities = [("gene", "BRCA2")]
        query = self.dataset.query(
            entities=entities,
            target_fields=[dict(name="user", expression='user()["email"]')],
            limit=1,
        )
        self.assertEqual(query.count(), 1)
        for record in query:
            assert "@quartzbio.com" in record["user"]

    def test_join(self):
        query_a = self.dataset2.query(fields=["gene"], limit=1).filter(gene="MAN2B1")
        query_b = self.dataset2.query(fields=["gene", "variant"])
        join_query = query_a.join(query_b, key="gene")

        count = 0
        for row in join_query:
            self.assertTrue("b_gene" not in row)
            self.assertEqual(row["gene"], "MAN2B1")
            count += 1
            if count == 10:
                break

    def test_join_custom_prefix(self):
        query_a = self.dataset2.query(fields=["gene", "variant"], limit=1).filter(
            gene="MAN2B1"
        )
        query_b = self.dataset2.query(fields=["gene", "variant"])
        join_query = query_a.join(query_b, key="gene", prefix="query_b_")

        count = 0
        for row in join_query:
            self.assertTrue("query_b_variant" in row)
            count += 1
            if count == 10:
                break

    def test_join_enable_always_prefix(self):
        query_a = self.dataset2.query(fields=["gene"], limit=1).filter(gene="MAN2B1")
        query_b = self.dataset2.query(fields=["gene", "variant"])
        join_query = query_a.join(query_b, key="gene", always_prefix=True)

        count = 0
        for row in join_query:
            self.assertTrue("b_variant" in row)
            count += 1
            if count == 10:
                break

    def test_join_empty_query_a(self):
        # Empty query_a
        query_a = self.dataset2.query(fields=["gene"]).filter(gene="PoshRoyalGene")
        query_b = self.dataset2.query(fields=["gene"])
        join_query = query_a.join(query_b, key="gene")

        self.assertEqual(list(join_query), [])

    def test_join_empty_query_b(self):
        query_a = self.dataset2.query(fields=["gene"], limit=5).filter(gene="MAN2B1")
        # Empty query_b
        query_b = self.dataset2.query(fields=["gene"]).filter(gene="PoshRoyalGene")
        join_query = query_a.join(query_b, key="gene")

        self.assertEqual(len(join_query), 5)

        for row in join_query:
            for key, value in row.items():
                if "b_" in key:
                    self.assertEqual(value, None)

    def test_join_multiple_join(self):
        query_a = self.dataset2.query(fields=["gene"], limit=1).filter(gene="MAN2B1")
        query_b = self.dataset2.query(fields=["gene"])
        query_c = self.dataset2.query(fields=["gene"])

        join_query = query_a.join(query_b, key="gene", prefix="b_")
        multiple_join_query = join_query.join(query_c, key="gene", prefix="c_")

        for record in multiple_join_query:
            self.assertTrue("gene" in record)
            self.assertTrue("b_gene" in record)
            self.assertTrue("c_gene" in record)

    def test_join_with_list_key(self):
        """Test a join where the key is a list. In this case, clinical_significance is a list.
        Ensure that the output contains a single value in each resulting record."""

        annotator_params = {
            "pre_annotation_expression": "explode(record, ['clinical_significance'])"
        }
        query_a = (
            self.dataset2.query(
                fields=["clinical_significance", "variant"],
                annotator_params=annotator_params,
            )
            .filter(gene="MAN2B1")
            .limit(10)
        )
        query_b = (
            self.dataset2.query(fields=["clinical_significance", "gene"])
            .filter(gene="MAN2B1")
            .limit(10)
        )

        for i in query_a.join(query_b, "clinical_significance"):
            self.assertFalse(isinstance(i["clinical_significance"], list))
            self.assertTrue("_errors" not in i)

    def test_join_with_list_values(self):
        """Test a join where one of the fields is a list.
        In this case, clinical_significance is a list.
        Ensure that the output contains a list of strings (not lists)."""

        annotator_params = {
            "pre_annotation_expression": "explode(record, ['clinical_significance'])"
        }
        query_a = (
            self.dataset2.query(
                fields=["clinical_significance", "variant"],
                annotator_params=annotator_params,
            )
            .filter(gene="MAN2B1")
            .limit(10)
        )
        query_b = (
            self.dataset2.query(fields=["clinical_significance", "gene", "variant"])
            .filter(gene="MAN2B1")
            .limit(10)
        )

        for i in query_a.join(query_b, "variant"):
            # Since each value contains one element, the resulting output is just a flat string
            # as the API takes the first value in the list.
            self.assertFalse(isinstance(i["clinical_significance"], list))
            self.assertFalse(isinstance(i["b_clinical_significance"], list))
            self.assertTrue("_errors" not in i)

    def test_join_pagination(self):
        # 50 records total
        query_a = self.dataset2.query(fields=["gene"], limit=50, page_size=10).filter(
            gene="MAN2B1"
        )
        # 367 records total which have gene='MAN2B1'
        query_b = self.dataset2.query(fields=["gene"])

        join_query = query_a.join(query_b, key="gene", prefix="b_")

        self.assertEqual(len(query_a), 50)
        self.assertEqual(len(query_b.filter(gene="MAN2B1")), 367)
        self.assertEqual(len(list(join_query)), 50 * 367)

    def test_join_query_with_key_b(self):
        query_a = self.dataset.query(
            fields=["agr", "gene_symbol"], limit=50, page_size=10
        )
        query_b = self.dataset2.query(fields=["gene", "variant"])
        join_query = query_a.join(
            query_b, key="gene", key_b="variant", always_prefix=True
        )
        for record in join_query:
            self.assertTrue("agr" in record)
            self.assertTrue("gene_symbol" in record)
            self.assertTrue("b_gene" in record)
            self.assertTrue("b_variant" in record)

    @unittest.skip(
        "Skip because API Host on GH pipelines doesn't host file at TEST_LARGE_TSV_FULL_PATH"
    )
    def test_query_large_file_into_dataframe(self):
        import pandas as pd

        expected_num_rows = 1015

        try:
            large_object = self.client.Object.get_by_full_path(
                self.TEST_LARGE_TSV_FULL_PATH
            )
            query = large_object.query()
            dataframe = pd.DataFrame(query)
        except Exception as e:
            self.fail("Exception {} was raised while querying large object".format(e))
        self.assertTrue(not dataframe.empty)
        self.assertEqual(expected_num_rows, len(dataframe))



import dash
import dash_html_components as html
import dash_core_components as dcc
import flask

from quartzbio.contrib.dash import QuartzBioDash


# Initialize the Dash app with QuartzBio auth.
app = QuartzBioDash(
    name=__name__,
    title="Example Dash App",
    app_url="http://localhost:8050",
    client_id="zxpjuufs7k65f26zyq3vq4hqqw2dmu13x77p7qg2",
    quartzbio_url="http://my.quartzbio.com",
)


def current_user():
    if app.auth:
        user = flask.g.client.User.retrieve()
        return [
            html.Div(children="Logged-in as: {}".format(user.full_name)),
            html.A("Log out", href="/_dash-logout"),
        ]
    else:
        return [html.P("(QuartzBio Auth not configured)")]


def layout():
    return html.Div(
        [
            html.H1("Welcome to your Dash+QuartzBio app"),
            html.P(current_user()),
            dcc.Dropdown(
                id="dropdown",
                options=[{"label": i, "value": i} for i in ["A", "B"]],
                value="A",
            ),
            dcc.Graph(id="graph"),
        ],
        className="container",
    )


app.layout = html.Div(
    [dcc.Location(id="url", refresh=False), html.Div(id="page-content")]
)


#
# Your Dash app callback functions
#


@app.callback(
    dash.dependencies.Output("page-content", "children"),
    [dash.dependencies.Input("url", "pathname")],
)
def display_page(pathname):
    """Render the layout depending on the pathname."""
    return layout()


@app.callback(
    dash.dependencies.Output("graph", "figure"),
    [dash.dependencies.Input("dropdown", "value")],
)
def update_graph(dropdown_value):
    return {
        "layout": {
            "title": "Graph of {}".format(dropdown_value),
            "margin": {"l": 20, "b": 20, "r": 10, "t": 60},
        },
        "data": [{"x": [1, 2, 3], "y": [4, 1, 2]}],
    }


app.css.append_css({"external_url": "https://codepen.io/chriddyp/pen/bWLwgP.css"})

if __name__ == "__main__":
    app.run_server(debug=True)



import unittest

from quartzbio.resource.util import class_to_api_name


class ConversionTest(unittest.TestCase):
    def test_class_to_api_name(self):
        for class_name, expect in [
            ("Annotation", "annotations"),
            ("DataField", "data_fields"),
            ("Depository", "depositories"),
        ]:
            self.assertEqual(class_to_api_name(class_name), expect)



# -*- coding: utf-8 -*-

import contextlib
import sys
import unittest
import quartzbio
import quartzbio.cli.auth as auth


@contextlib.contextmanager
def nostdout():
    savestderr = sys.stdout

    class Devnull(object):
        def write(self, _):
            pass

    sys.stdout = Devnull()
    try:
        yield
    finally:
        sys.stdout = savestderr


class TestLogin(unittest.TestCase):
    def setUp(self):
        self.api_host = quartzbio.get_api_host()
        # temporarily replace with dummy methods for testing
        self.delete_credentials = auth.delete_credentials
        auth.delete_credentials = lambda: None

    def tearDown(self):
        quartzbio.client._host = self.api_host
        auth.delete_credentials = self.delete_credentials

    def test_bad_login(self):
        with nostdout():
            self.assertEqual(auth.login_and_save_credentials(), None, "Invalid login")

            # Test invalid host
            quartzbio.client._host = "https://some.fake.domain.foobar"
            self.assertEqual(auth.login_and_save_credentials(), None, "Invalid login")

    def test_init_login(self):
        from quartzbio import login
        from quartzbio.client import client

        _auth = client._auth

        client._auth = None
        login(api_key="TEST_KEY")
        self.assertEqual(client._auth.token, "TEST_KEY")

        # Reset the key
        client._auth = _auth




from .helper import QuartzBioTestCase
from quartzbio.errors import QuartzBioError
from quartzbio.resource import DatasetImport
from quartzbio.client import client


class ErrorTests(QuartzBioTestCase):
    def test_request_error(self):
        try:
            # two errors get raised
            DatasetImport.create(
                dataset_id="510113719950913753",
                manifest=dict(files=[dict(filename="soemthing.md")]),
            )
        except QuartzBioError as e:
            self.assertTrue("Error (dataset_id):" in str(e), e)
            self.assertTrue("Invalid dataset" in str(e), e)
            self.assertTrue("Error (manifest):" in str(e), e)
            self.assertTrue("Each file must" in str(e), e)


class ErrorTestsAuth(QuartzBioTestCase):
    def test_no_body(self):
        # Remove auth
        auth = client._auth
        client._auth = None
        try:
            client.whoami()
        except QuartzBioError as e:
            self.assertTrue("Error: Authentication credentials were not" in str(e), e)

        client._auth = auth



from .apiresource import ListableAPIResource
from .apiresource import CreateableAPIResource
from .quartzbio_object import convert_to_quartzbio_object
from .task import Task

import time


class DatasetRestoreTask(CreateableAPIResource, ListableAPIResource):
    """
    DatasetRestoreTask represents the task to restore an archived dataset
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("dataset_id", "Dataset"),
        ("vault_id", "Vault"),
        ("status", "Status"),
        ("created_at", "Created"),
    )

    @property
    def dataset(self):
        response = self._client.get(self["dataset"]["url"], {})
        return convert_to_quartzbio_object(response, client=self._client)

    def follow(self, loop=True, sleep_seconds=Task.SLEEP_WAIT_DEFAULT):
        if self.status == "queued":
            print(
                "Waiting for Dataset Restore task (id = {0}) to start...".format(
                    self.id
                )
            )

        status = self.status
        while self.status in ["queued", "running"]:
            if self.status != status:
                print("Restore is now {0} (was {1})".format(self.status, status))
                status = self.status

            if self.status == "running":
                progress = self.metadata.get("progress", {}).get("restore")
                if progress:
                    print(
                        "Restore '{0}' is {1} (Progress: {2}%% completed)".format(
                            self.id, self.status, progress
                        )
                    )
                else:
                    print("Restore '{0}' is {1}".format(self.id, self.status))

            if not loop:
                return

            time.sleep(sleep_seconds)
            self.refresh()

        if self.status == "completed":
            print("Restore complete! Dataset is now available")




import os
import unittest

import quartzbio


class QuartzBioTestCase(unittest.TestCase):
    # 'quartzbio:public:/HGNC/3.3.0-2020-10-29/HGNC'
    TEST_DATASET_FULL_PATH = "quartzbio:Public:/HGNC/3.3.1-2021-08-25/HGNC"
    # 'quartzbio:public:/ClinVar/5.1.0-20200720/Variants-GRCH38'
    TEST_DATASET_FULL_PATH_2 = (
        "quartzbio:Public:/ClinVar/5.2.0-20210110/Variants-GRCH38"
    )
    # 'quartzbio:public:/HGNC/3.3.0-2020-10-29/hgnc_1000_rows.txt'
    TEST_FILE_FULL_PATH = "quartzbio:Public:/HGNC/3.3.1-2021-08-25/HGNC-3-3-1-2021-08-25-HGNC-1904014068027535892-20230418174248.json.gz"  # noqa
    TEST_LARGE_TSV_FULL_PATH = ""

    def setUp(self):
        super(QuartzBioTestCase, self).setUp()
        api_key = os.environ.get("QUARTZBIO_API_KEY", None)
        api_host = os.environ.get("QUARTZBIO_API_HOST", None)
        self.client = quartzbio.QuartzBioClient(host=api_host, token=api_key)

    def check_response(self, response, expect, msg):
        subset = [(key, response[key]) for key in [x[0] for x in expect]]
        self.assertEqual(subset, expect)



# -*- coding: utf-8 -*-

from collections import namedtuple

import quartzbio

from netrc import netrc as _netrc, NetrcParseError
import os


def as_netrc_machine(api_host):
    return api_host.replace("https://", "").replace("http://", "")


class netrc(_netrc):
    """
    Adds a save() method to netrc
    """

    @staticmethod
    def path():
        if os.name == "nt":
            # Windows
            path = "~\\_quartzbio\\credentials"
        else:
            # *nix
            path = "~/.quartzbio/credentials"

        try:
            path = os.path.expanduser(path)
        except KeyError:
            # os.path.expanduser can fail when $HOME is undefined and
            # getpwuid fails. See http://bugs.python.org/issue20164
            raise IOError("Could not find any home directory for '{0}'".format(path))

        if not os.path.isdir(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        # create an empty credentials file if it doesn't exist
        if not os.path.exists(path):
            try:
                open(path, "a").close()
            except IOError:
                raise Exception(
                    "Could not create a QuartzBio credentials file"
                    " at '%s', permission denied." % path
                )
        return path

    def save(self, path):
        """Dump the class data in the format of a .netrc file."""
        rep = ""
        for host in self.hosts.keys():
            attrs = self.hosts[host]
            rep = (
                rep + "machine " + host + "\n\tlogin " + str(attrs[0]) + "\n"
            )
            if attrs[1]:
                rep = rep + "\taccount " + str(attrs[1]) + "\n"
            rep = rep + "\tpassword " + str(attrs[2]) + "\n"

        f = open(path, "w")
        f.write(rep)
        f.close()


class CredentialsError(BaseException):
    """
    Raised if the credentials are not found.
    """

    pass


ApiCredentials = namedtuple(
    "ApiCredentials", ["api_host", "email", "token_type", "token"]
)


def get_credentials(api_host: str = None) -> ApiCredentials:
    """
    Returns the user's stored API key if a valid credentials file is found.
    Raises CredentialsError if no valid credentials file is found.
    """

    try:
        netrc_obj = netrc(netrc.path())
        if not netrc_obj.hosts:
            return None
    except (IOError, TypeError, NetrcParseError) as e:
        raise CredentialsError("Could not open credentials file: " + str(e))

    netrc_host: str = None

    # if user provides a host, then find its token in the credentials file
    if api_host is not None:
        api_host = api_host.removeprefix("https://")
        if api_host in netrc_obj.hosts:
            netrc_host = api_host
        else:
            # login has failed for the requested host,
            # the rest of the credentials file is ignored
            return None
    else:
        # If there are no stored credentials for the default host,
        # but there are other stored credentials, use the first
        # available option that ends with '.api.quartzbio.com',
        netrc_host = next(
            filter(lambda h: h.endswith(".api.quartzbio.com"), netrc_obj.hosts), None
        )

    # Otherwise use the first available.
    if netrc_host is None:
        netrc_host = next(iter(netrc_obj.hosts))

    if netrc_host is not None:
        return ApiCredentials(
            "https://" + netrc_host, *netrc_obj.authenticators(netrc_host)
        )
    return None

    # api_host_no_schema = as_netrc_machine(api_host)
    # if api_host_no_schema in netrc_obj.hosts:
    #     proto = urlparse(api_host).scheme
    #     return (proto + api_host_no_schema,) + netrc_obj.authenticators(api_host_no_schema)

    # If the preferred host is not the global default, don't try
    # to select any other.
    # if api_host_no_schema != "api.quartzbio.com":
    #     return None


def delete_credentials():
    try:
        netrc_path = netrc.path()
        rc = netrc(netrc_path)
    except (IOError, TypeError, NetrcParseError) as e:
        raise CredentialsError("Could not open netrc file: " + str(e))

    try:
        del rc.hosts[as_netrc_machine(quartzbio.get_api_host())]
    except KeyError:
        pass
    else:
        rc.save(netrc_path)


def save_credentials(email, token, token_type="Token", api_host=None):
    api_host = api_host or quartzbio.get_api_host()

    try:
        netrc_path = netrc.path()
        rc = netrc(netrc_path)
    except (IOError, TypeError, NetrcParseError) as e:
        raise CredentialsError("Could not open netrc file: " + str(e))

    # Overwrites any existing credentials
    rc.hosts[as_netrc_machine(api_host)] = (email, token_type, token)
    rc.save(netrc_path)



# -*- coding: utf-8 -*-

import os
import sys
import subprocess
import locale
import logging

logger = logging.getLogger("quartzbio")

try:
    # reload() for Python3
    from importlib import reload
except ImportError:
    pass

std_handles = [sys.stdin, sys.stdout, sys.stderr]
try:
    # Switch from the default input ASCII encoding to the default locale.
    # The Python runtime will use this when it has to decode a
    # string buffer to unicode. This is not needed in Python3.

    # However reload(sys), used below resets stdin, stdout, and stderr
    # which is bad if they've already been reassigned. An ipython
    # notebook shell, for example, sets up its own stdout.
    # See GitHub issue #43 and #21.
    reload(sys).setdefaultencoding(locale.getdefaultlocale()[1])
    locale.setlocale(locale.LC_ALL, "")
except:
    pass
finally:
    sys.stdin, sys.stdout, sys.stderr = std_handles


# Set rows and columns and colors


def set_from_env(name, default_value):
    try:
        return int(os.environ[name])
    except:
        return default_value


TTY_ROWS = set_from_env("LINES", 24)
TTY_COLS = set_from_env("COLUMNS", 80)

TTY_COLORS = True

if sys.stdout.isatty():
    try:
        with open(os.devnull, "w") as fnull:
            rows, cols = subprocess.check_output(["stty", "size"], stderr=fnull).split()
            TTY_ROWS = int(rows)
            TTY_COLS = int(cols)
    except:
        logger.warn(
            "Cannot detect terminal column width.\nUsing value "
            "from environment variables and/or internal defaults."
        )
else:
    TTY_COLORS = False


def pretty_int(num):
    return locale.format_string("%d", int(num), grouping=True)


# Basic color support


def green(text):
    if not TTY_COLORS:
        return text
    return "\033[32m" + text + "\033[39m"


def red(text):
    if not TTY_COLORS:
        return text
    return "\033[31m" + text + "\033[39m"


def yellow(text):
    if not TTY_COLORS:
        return text
    return "\033[33m" + text + "\033[39m"


def blue(text):
    if not TTY_COLORS:
        return text
    return "\033[34m" + text + "\033[39m"


def pager(fn, **kwargs):
    try:
        import tty

        fd = sys.stdin.fileno()
        old = tty.tcgetattr(fd)
        tty.setcbreak(fd)

        def getchar():
            sys.stdin.read(1)

    except (ImportError, AttributeError):
        tty = None

        def getchar():
            sys.stdin.readline()[:-1][:1]

    try:
        page = 1
        res = fn(page=page, **kwargs)
        has_next = res.links["next"]
        sys.stdout.write(str(res) + "\n")

        while has_next:
            sys.stdout.write("-- More --")
            sys.stdout.flush()
            c = getchar()
            page += 1
            res = fn(page=page, **kwargs)
            has_next = res.links["next"]

            if c in ("q", "Q"):
                sys.stdout.write("\r          \r")
                break

            sys.stdout.write("\n" + str(res) + "\n")
    finally:
        if tty:
            tty.tcsetattr(fd, tty.TCSAFLUSH, old)




import unittest
import uuid
import os
import shutil

import requests

import mock

from .helper import QuartzBioTestCase
from quartzbio.test.client_mocks import fake_object_create, fake_object_save
from quartzbio.test.client_mocks import fake_dataset_create


def get_uuid_str():
    return str(uuid.uuid4())


class ObjectTests(QuartzBioTestCase):
    def test_object_paths(self):
        # List objects and get them, but only in the public vault
        # this way we minimize flaky tests.
        vaults = self.client.Vault.all()
        for vault in vaults:
            if not vault.is_public:
                continue

            for file_ in list(vault.ls().quartzbio_objects())[:5]:
                o_path, _ = self.client.Object.validate_full_path(file_.full_path)
                self.assertEqual(o_path, file_.full_path)

                # assert path is gettable
                self.client.Object.get_by_full_path(o_path)

        with self.assertRaises(Exception):
            self.client.Object.get_by_full_path("what/is/this")

    def test_object_output(self):
        case = "acme:myVault/uploads_folder"
        expected = "acme:myVault:/uploads_folder"
        p, path_dict = self.client.Object.validate_full_path(case)

        self.assertEqual(p, expected)
        self.assertEqual(path_dict["full_path"], expected)
        self.assertEqual(path_dict["path"], "/uploads_folder")
        self.assertEqual(path_dict["parent_path"], "/")
        self.assertEqual(path_dict["parent_full_path"], "acme:myVault:/")
        self.assertEqual(path_dict["filename"], "uploads_folder")
        self.assertEqual(path_dict["domain"], "acme")
        self.assertEqual(path_dict["vault"], "myVault")
        self.assertEqual(path_dict["vault_full_path"], "acme:myVault")

    def test_object_path_cases(self):
        user = self.client.User.retrieve()
        domain = user.account.domain
        user_vault = "{0}:user-{1}".format(domain, user.id)
        test_cases = [
            ["acme:myVault:/uploads_folder", "acme:myVault:/uploads_folder"],
            [
                "acme:myVault/folder1/project: ABCD",
                "acme:myVault:/folder1/project: ABCD",
            ],
            [
                "myVault/folder1/project: ABCD",
                "{0}:myVault:/folder1/project: ABCD".format(domain),
            ],
            [
                "myVault:/uploads_folder",
                "{0}:myVault:/uploads_folder".format(domain),
            ],  # noqa
            # New full path formats
            ["~/uploads_folder", "{0}:/uploads_folder".format(user_vault)],
            ["~/", "{0}:/".format(user_vault)],
            [
                "myVault/uploads_folder",
                "{0}:myVault:/uploads_folder".format(domain),
            ],  # noqa
            ["acme:myVault/uploads_folder", "acme:myVault:/uploads_folder"],
        ]
        for case, expected in test_cases:
            p, _ = self.client.Object.validate_full_path(case)
            self.assertEqual(p, expected)

        error_test_cases = [
            "",
            "/hello",
            "myVault",
            "oops:myDomain:myVault",
            "{0}:myVault".format(domain),
            "acme:myVault",
        ]
        for case in error_test_cases:
            with self.assertRaises(Exception):
                v, v_paths = self.client.Object.validate_full_path(case)

    def test_object_path_cases_with_overrides(self):
        user = self.client.User.retrieve()
        domain = user.account.domain
        user_vault = "{0}:user-{1}".format(domain, user.id)

        test_cases = [
            "acme:myVault:/folder",
            "myVault:/folder",
            "myVault/folder",
            "/folder",
        ]

        for case in test_cases:
            p, _ = self.client.Object.validate_full_path(case, vault="foobar")
            expected = "{0}:foobar:/folder".format(domain)
            self.assertEqual(p, expected)

            p, _ = self.client.Object.validate_full_path(case, vault="foo:bar")
            expected = "foo:bar:/folder"
            self.assertEqual(p, expected)

            p, _ = self.client.Object.validate_full_path(
                case, vault="foo:bar", path="/baz"
            )
            expected = "foo:bar:/baz"
            self.assertEqual(p, expected)

            p, _ = self.client.Object.validate_full_path(
                case, vault="foo:bar", path="/baz/bon"
            )
            expected = "foo:bar:/baz/bon"
            self.assertEqual(p, expected)

        # Test cases where just the path changes
        case = "acme:myVault:/folder"
        p, _ = self.client.Object.validate_full_path(case, path="foo/bar/baz")
        expected = "acme:myVault:/foo/bar/baz"
        self.assertEqual(p, expected)

        case = "~/folder"
        p, _ = self.client.Object.validate_full_path(case, path="foo/bar/baz")
        expected = "{0}:/foo/bar/baz".format(user_vault)
        self.assertEqual(p, expected)

    @mock.patch("quartzbio.resource.Object.create")
    def test_object_has_tag(self, ObjectMock):
        ObjectMock.side_effect = fake_object_create

        tags = ["foo", "bar", "BIZ"]
        obj = self.client.Object.create(name="blah", tags=tags)
        self.assertEqual(obj.tags, tags)
        self.assertTrue(obj.has_tag("FOO"))
        self.assertTrue(obj.has_tag("foo"))
        self.assertTrue(obj.has_tag("BAr"))
        self.assertTrue(obj.has_tag("BAr"))
        self.assertTrue(obj.has_tag("biz"))
        self.assertFalse(obj.has_tag("BAz"))
        self.assertFalse(obj.has_tag("baz"))

        obj = self.client.Object.create(name="blah_untagged")
        self.assertEqual(obj.tags, [])
        self.assertFalse(obj.has_tag("foo"))

    @mock.patch("quartzbio.resource.Dataset.create")
    @mock.patch("quartzbio.resource.Object.create")
    def test_object_dataset_getattr(self, ObjectCreate, DatasetCreate):
        ObjectCreate.side_effect = fake_object_create
        DatasetCreate.side_effect = fake_dataset_create

        valid_attrs = [
            "lookup",
            "beacon",
            "import_file",
            "export",
            "migrate",
            "fields",
            "template",
            "imports",
            "commits",
            "activity",
            "saved_queries",
        ]

        ds = self.client.Dataset.create(name="foo")
        ds_obj = self.client.Object.create(name="foo_dataset", object_type="dataset")
        file_ = self.client.Object.create(name="foo_file", object_type="file")
        folder_ = self.client.Object.create(name="foo_folder", object_type="folder")
        for attr in valid_attrs:
            self.assertTrue(getattr(ds, attr))
            self.assertTrue(getattr(ds_obj, attr))
            with self.assertRaises(AttributeError):
                getattr(file_, attr)
            with self.assertRaises(AttributeError):
                getattr(folder_, attr)

        # Test that any old attr doesnt work
        fake_attr = "foobar"
        for obj in [file_, folder_, ds, ds_obj]:
            with self.assertRaises(AttributeError):
                getattr(obj, fake_attr)

    @mock.patch("quartzbio.resource.Object.create")
    @mock.patch("quartzbio.resource.Object.save")
    def test_object_remove_tag(self, ObjectCreate, ObjectSave):
        ObjectCreate.side_effect = fake_object_create
        ObjectSave.side_effect = fake_object_save

        tags = ["tag1", "tag2", "tag3"]
        file_ = self.client.Object.create(
            filename="foo_file", object_type="file", tags=tags
        )
        for tag in tags:
            self.assertTrue(file_.has_tag(tag))

        tags_for_removal = ["tag1", "tag2"]
        file_.tag(tags=tags_for_removal, remove=True, apply_save=True)

        # test that given tags are removed
        for tag in tags_for_removal:
            self.assertFalse(file_.has_tag(tag))

        updated_tags = [tag for tag in tags if tag not in tags_for_removal]

        # test that tags which have not been removed are still there
        for tag in updated_tags:
            self.assertTrue(file_.has_tag(tag))

    @mock.patch("quartzbio.resource.Object.create")
    @mock.patch("quartzbio.resource.Object.save")
    def test_object_untag(self, ObjectCreate, ObjectSave):
        ObjectCreate.side_effect = fake_object_create
        ObjectSave.side_effect = fake_object_save

        tags = ["tag1", "tag2", "tag3"]
        file_ = self.client.Object.create(
            filename="foo_file", object_type="file", tags=tags
        )
        for tag in tags:
            self.assertTrue(file_.has_tag(tag))

        tags_for_untagging = ["tag1", "tag2"]
        file_.untag(tags=tags_for_untagging, apply_save=True)

        # test that given tags are untagged
        for tag in tags_for_untagging:
            self.assertFalse(file_.has_tag(tag))

        updated_tags = [tag for tag in tags if tag not in tags_for_untagging]

        # test that tags which have not been untagged are still there
        for tag in updated_tags:
            self.assertTrue(file_.has_tag(tag))

    @mock.patch("quartzbio.resource.Object.create")
    @mock.patch("quartzbio.resource.Object.save")
    def test_object_add_tag_non_iterable_or_string(self, ObjectCreate, ObjectSave):
        ObjectCreate.side_effect = fake_object_create
        ObjectSave.side_effect = fake_object_save

        # test that a string is added propery
        tags = "tag1"
        file_ = self.client.Object.create(filename="foo_file", object_type="file")
        file_.tag(tags)
        self.assertTrue(file_.has_tag(tags))

        # test that non iterable (e.g. integer) added properly
        tags = 1
        file_.tag(tags)
        self.assertTrue(file_.has_tag(tags))

    def test_object_set_metadata(self):
        folder = self.client.Object.get_or_create_by_full_path(
            "~/{}".format(uuid.uuid4()), object_type="folder"
        )
        metadata = folder.metadata

        metadata["foo_1"] = "bar_1"
        folder.save()

        # Test that item has been set successfully
        self.assertTrue(metadata.get("foo_1") == "bar_1")

        metadata = {"foo_2": "bar_2"}
        self.assertTrue(metadata.get("foo_2") == "bar_2")

        # Test that direct assignement overrides QuartzBio object type with the given dictionary
        self.assertTrue(
            len(metadata) == 1
            and metadata.pop("foo_2") == "bar_2"
            and len(metadata) == 0
        )

        folder.delete(force=True)

    def test_object_set_metadata_empty_dict_list(self):
        folder = self.client.Object.get_or_create_by_full_path(
            "~/{}".format(uuid.uuid4()), object_type="folder"
        )

        folder.metadata["foo"] = "bar"
        folder.save()

        folder.metadata = {}
        folder.save()

        # Test that 'metadata' is an empty dict
        self.assertTrue(folder.metadata == {})

        folder.metadata = []
        folder.save()

        # Test that 'metadata' is an empty dict (can't be a list)
        self.assertTrue(folder.metadata == {})

        folder.delete(force=True)

    def test_object_serialize_metadata(self):
        tpl = self.client.DatasetTemplate()
        tpl.name = "test"
        res = tpl.serialize(tpl)
        self.assertEqual(res, {"name": "test"})

        tpl.metadata = {"hello": "world"}
        res = tpl.serialize(tpl)
        self.assertEqual(res, {"metadata": {"hello": "world"}, "name": "test"})

    @mock.patch("quartzbio.resource.Object.create")
    @mock.patch("quartzbio.QuartzBioClient.post")
    def test_object_query(self, QuartzBioClientPost, ObjectCreate):
        ObjectCreate.side_effect = fake_object_create
        valid_response = {"took": 100, "total": 2, "results": [{"foo": 1}, {"bar": 2}]}

        file = self.client.Object.create(
            filename="foo_file.json", object_type="file", size=100
        )
        from .. import QuartzBioClient

        file._client = QuartzBioClient()

        QuartzBioClientPost.return_value = valid_response
        resp = file.query()

        self.assertEqual(list(resp), valid_response["results"])

    @unittest.skip("Skip because API Host on GH pipelines doesn't support versioning.")
    def test_object_versioning(self):
        vault = self.client.Vault.get_personal_vault()
        initial_versioning_status = vault["versioning"]
        vault.enable_versioning()

        file = self.client.Object.create(
            filename="quartzbio-py-test.txt", object_type="file", vault_id=vault.id
        )

        self.addCleanup(
            ObjectTests.clean_up_after_object_versioning,
            vault,
            file,
            initial_versioning_status,
        )

        assert file["version_count"] == 1

        file = self.client.Object.create(
            filename="quartzbio-py-test.txt", object_type="file", vault_id=vault.id
        )

        assert file["version_count"] == 2

        versions = file.list_versions()
        assert len(versions["data"]) == 2

        version_id_1 = versions["data"][0]["id"]

        file.delete_version(version_id_1)
        versions = file.list_versions(include_deleted=False)
        assert len(versions["data"]) == 1

        versions = file.list_versions()
        assert versions["data"][0]["is_deleted"]

        file.undelete_version(version_id_1)
        versions = file.list_versions()
        assert not versions["data"][0]["is_deleted"]

    @staticmethod
    def clean_up_after_object_versioning(vault, file, versioning_status):
        from .test_vault import VaultTests

        file.delete(force=True)
        VaultTests.clean_up_after_vault_versioning(vault, versioning_status)

    @unittest.skip("Skip because API Host on GH pipelines doesn't support shortcuts.")
    def test_shortcuts(self):
        vault = self.client.Vault.get_personal_vault()

        folder_name = "{}-test-folder".format(str(uuid.uuid4()))
        folder = self.client.Object.create(
            filename=folder_name, object_type="folder", vault_id=vault.id
        )

        file = self.client.Object.create(
            filename="shorcut-test.txt", object_type="file", vault_id=vault.id
        )

        shortcut_folder_full_path = (
            vault.full_path + ":/{}-test-shortcut-folder".format(get_uuid_str())
        )
        shortcut_folder = folder.create_shortcut(shortcut_folder_full_path)
        self.assertTrue(shortcut_folder.is_shortcut)
        self.assertEqual(shortcut_folder.get_target().id, folder.id)

        shortcut_file_full_path = vault.full_path + ":/{}-test-shortcut-file".format(
            get_uuid_str()
        )
        shortcut_file = file.create_shortcut(shortcut_file_full_path)
        self.assertTrue(shortcut_file.is_shortcut)
        self.assertEqual(shortcut_file.get_target().id, file.id)

        shortcut_url_name = "{}-test-shortcut-url".format(get_uuid_str())
        shortcut_url = self.client.Object.create(
            filename=shortcut_url_name,
            object_type="shortcut",
            vault_id=vault.id,
            target={
                "url": "www.google.com",
                "object_type": "url",
            },
        )
        self.assertTrue(shortcut_url.is_shortcut)
        self.assertEqual(shortcut_url.get_target(), "www.google.com")


class ObjectUploadTests(QuartzBioTestCase):
    def setUp(self):
        super(ObjectUploadTests, self).setUp()
        self.tempdir = "/tmp/quartzbio-tests"
        if not os.path.exists(self.tempdir):
            os.makedirs(self.tempdir)

        # Do not delete vault at end of
        # test as other tests run in
        # parallel and may expect files
        vault_name = "ObjectUploadTests"
        self.vault = self.client.Vault.get_or_create_by_full_path(vault_name)

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    @mock.patch("quartzbio.resource.Dataset.create")
    @mock.patch("quartzbio.resource.Object.create")
    def test_archive_non_files(self, DatasetCreate, ObjectCreate):
        DatasetCreate.side_effect = fake_object_create
        ObjectCreate.side_effect = fake_object_create

        folder_object = self.client.Object.create(
            name="foo_folder", object_type="folder"
        )
        with self.assertRaises(NotImplementedError):
            folder_object._archive("archive_folder")

        dataset_object = self.client.Object.create(
            name="foo_dataset", object_type="dataset"
        )
        with self.assertRaises(NotImplementedError):
            dataset_object._archive("archive_folder")

    def test_create_folder(self):
        folder_full_path = self.vault.full_path + ":/{}-test-folder".format(
            get_uuid_str()
        )
        self.client.Object.create_folder(self.vault, folder_full_path)
        obj = self.client.Object.get_by_full_path(folder_full_path)
        self.assertEqual(obj.object_type, "folder")

    def test_upload_file(self):
        local_path = os.path.join(self.tempdir, "{}-file.txt".format(get_uuid_str()))
        with open(local_path, "w") as fp:
            fp.write("sample file")
        remote_path = "/"

        self.client.Object.upload_file(local_path, remote_path, self.vault.full_path)

        expected_remote_path = (
            self.vault.full_path + ":/" + os.path.basename(local_path)
        )
        obj = self.client.Object.get_by_full_path(expected_remote_path)
        self.assertEqual(obj.full_path, expected_remote_path)

        download_url = obj.download_url()
        response = requests.request(method="get", url=download_url)
        self.assertEqual(response.content.decode("utf-8"), "sample file")

    def test_skip_md5_match(self):
        local_path = os.path.join(self.tempdir, "{}-file.txt".format(get_uuid_str()))
        with open(local_path, "w") as fp:
            fp.write("sample file")
        remote_path = "/"
        self.client.Object.upload_file(local_path, remote_path, self.vault.full_path)

        with mock.patch("quartzbio.resource.Object.create") as ObjectCreate:
            self.client.Object.upload_file(
                local_path, remote_path, self.vault.full_path
            )
            self.assertEqual(ObjectCreate.call_count, 0)

    @mock.patch("quartzbio.resource.object.Object._get_timestamp")
    def test_archive_if_updated(self, TimeStamp):
        filename = "{}-file.txt.gz".format(get_uuid_str())
        filename_base = filename.split(".")[0]
        local_path = os.path.join(self.tempdir, filename)
        with open(local_path, "w") as fp:
            fp.write("sample file")
        remote_path = "/"
        self.client.Object.upload_file(local_path, remote_path, self.vault.full_path)

        with open(local_path, "w") as fp:
            fp.write("sample file updated")

        TimeStamp.return_value = "timestamp_string"

        self.client.Object.upload_file(
            local_path, remote_path, self.vault.full_path, archive_folder="archive"
        )

        expected_archive_dir = (
            self.vault.full_path
            + ":/"
            + os.path.join("archive", remote_path.lstrip("/"))
        )
        dir_object = self.client.Object.get_by_full_path(expected_archive_dir)
        self.assertEqual(dir_object.object_type, "folder")

        expected_archive_full_path = (
            self.vault.full_path
            + ":/archive/{}_timestamp_string.txt.gz".format(filename_base)
        )
        archive_obj = self.client.Object.get_by_full_path(expected_archive_full_path)
        download_url = archive_obj.download_url()
        response = requests.request(method="get", url=download_url)
        self.assertEqual(response.content.decode("utf-8"), "sample file")

        expected_new_path = self.vault.full_path + ":/" + filename
        new_obj = self.client.Object.get_by_full_path(expected_new_path)
        download_url = new_obj.download_url()
        response = requests.request(method="get", url=download_url)
        self.assertEqual(response.content.decode("utf-8"), "sample file updated")



import time

from .apiresource import ListableAPIResource
from .apiresource import CreateableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import DownloadableAPIResource
from .quartzbio_object import convert_to_quartzbio_object
from .task import Task


def follow_commits(task, sleep_seconds):
    """Utility used to wait for commits"""
    while True:
        unfinished_commits = [
            c for c in task.dataset_commits if c.status in ["queued", "running"]
        ]

        if not unfinished_commits:
            print("All commits have finished processing")
            break

        print(
            "{0}/{1} commits have finished processing".format(
                len(unfinished_commits), len(task.dataset_commits)
            )
        )

        # Prints a status for each one
        for commit in unfinished_commits:
            commit.follow(loop=False, sleep_seconds=sleep_seconds)

        time.sleep(sleep_seconds)

        # refresh Task to get fresh dataset commits
        task.refresh()


class DatasetCommit(CreateableAPIResource, ListableAPIResource, UpdateableAPIResource, DownloadableAPIResource):
    """
    DatasetCommits represent a change made to a Dataset.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("title", "Title"),
        ("description", "Description"),
        ("status", "Status"),
        ("created_at", "Created"),
    )

    @property
    def dataset(self):
        return convert_to_quartzbio_object(self["dataset"], client=self._client)

    @property
    def parent_object(self):
        """Get the commit objects parent Import or Migration"""
        from . import types

        parent_klass = types.get(self.parent_job_model.split(".")[1])
        return parent_klass.retrieve(self.parent_job_id, client=self._client)

    def follow(self, loop=True, sleep_seconds=Task.SLEEP_WAIT_DEFAULT):
        # Follow unfinished commits
        while self.status in ["queued", "running"]:
            if self.status == "running":
                print(
                    "Commit {3} is {0}: {1}/{2} records indexed".format(
                        self.status, self.records_modified, self.records_total, self.id
                    )
                )
            else:
                print("Commit {0} is {1}".format(self.id, self.status))

            # When following a parent DatasetImport we do not want to
            # loop for status updates. It will handle its own looping
            # so break out of loop and return here.
            if not loop:
                break

            # sleep
            time.sleep(sleep_seconds)

            # refresh status
            self.refresh()

        if loop:
            print("Commit '{0}' ({1}) is {2}".format(self.title, self.id, self.status))
            print(
                "View your imported data: " "https://my.quartzbio.com/data/{0}".format(
                    self["dataset"]["id"]
                )
            )



"""QuartzBio Application API Resource"""

from .apiresource import CreateableAPIResource
from .apiresource import ListableAPIResource
from .apiresource import SearchableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import DeletableAPIResource


class Application(
    CreateableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    SearchableAPIResource,
    UpdateableAPIResource,
):
    ID_ATTR = "client_id"
    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("client_id", "Client ID"),
        ("name", "Name"),
        ("description", "Description"),
        ("web_url", "web_url"),
    )
